import numpy as np
import pandas as pd
import xarray as xr
from scipy.interpolate import interpn

from foxes.core import States
from foxes.utils import wd2uv, uv2wd
from foxes.data import STATES, StaticData
import foxes.variables as FV
import foxes.constants as FC


class FieldDataNC(States):
    """
    Heterogeneous ambient states on a regular
    horizontal grid in NetCDF format.

    Attributes
    ----------
    data_source: str or xarray.Dataset
        The data or the file search pattern, should end with
        suffix '.nc'. One or many files.
    ovars: list of str
        The output variables
    var2ncvar: dict
        Mapping from variable names to variable names
        in the nc file
    fixed_vars: dict
        Uniform values for output variables, instead
        of reading from data
    states_coord: str
        The states coordinate name in the data
    x_coord: str
        The x coordinate name in the data
    y_coord: str
        The y coordinate name in the data
    h_coord: str
        The height coordinate name in the data
    pre_load: bool
        Flag for loading all data into memory during
        initialization
    weight_ncvar: str
        Name of the weight data variable in the nc file(s)
    bounds_error: bool
        Flag for raising errors if bounds are exceeded
    fill_value: number
        Fill value in case of exceeding bounds, if no bounds error
    time_format: str
        The datetime parsing format string
    interp_nans: bool
        Linearly interpolate nan values
    interpn_pars: dict, optional
        Additional parameters for scipy.interpolate.interpn

    :group: input.states

    """

    def __init__(
        self,
        data_source,
        output_vars,
        var2ncvar={},
        fixed_vars={},
        states_coord="Time",
        x_coord="UTMX",
        y_coord="UTMY",
        h_coord="height",
        pre_load=True,
        weight_ncvar=None,
        time_format="%Y-%m-%d_%H:%M:%S",
        sel=None,
        isel=None,
        interp_nans=False,
        verbosity=1,
        **interpn_pars,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or xarray.Dataset
            The data or the file search pattern, should end with
            suffix '.nc'. One or many files.
        output_vars: list of str
            The output variables
        var2ncvar: dict, optional
            Mapping from variable names to variable names
            in the nc file
        fixed_vars: dict, optional
            Uniform values for output variables, instead
            of reading from data
        states_coord: str
            The states coordinate name in the data
        x_coord: str
            The x coordinate name in the data
        y_coord: str
            The y coordinate name in the data
        h_coord: str
            The height coordinate name in the data
        pre_load: bool
            Flag for loading all data into memory during
            initialization
        weight_ncvar: str, optional
            Name of the weight data variable in the nc file(s)
        time_format: str
            The datetime parsing format string
        sel: dict, optional
            Subset selection via xr.Dataset.sel()
        isel: dict, optional
            Subset selection via xr.Dataset.isel()
        interp_nans: bool
            Linearly interpolate nan values
        verbosity: int
            Verbosity level for pre_load file reading
        interpn_pars: dict, optional
            Additional parameters for scipy.interpolate.interpn

        """
        super().__init__()

        self.data_source = data_source
        self.states_coord = states_coord
        self.ovars = output_vars
        self.fixed_vars = fixed_vars
        self.x_coord = x_coord
        self.y_coord = y_coord
        self.h_coord = h_coord
        self.weight_ncvar = weight_ncvar
        self.pre_load = pre_load
        self.time_format = time_format
        self.sel = sel
        self.isel = isel
        self.interpn_pars = interpn_pars
        self.interp_nans = interp_nans

        self.var2ncvar = {
            v: var2ncvar.get(v, v) for v in output_vars if v not in fixed_vars
        }

        self._inds = None
        self._N = None
        self._weights = None

        # pre-load file reading, usually prior to DaskRunner:
        if not isinstance(self.data_source, xr.Dataset):
            if "*" in str(self.data_source):
                pass
            else:
                self.data_source = StaticData().get_file_path(
                    STATES, self.data_source, check_raw=True
                )
            if verbosity:
                if pre_load:
                    print(
                        f"States '{self.name}': Reading data from '{self.data_source}'"
                    )
                else:
                    print(
                        f"States '{self.name}': Reading index from '{self.data_source}'"
                    )

            with xr.open_mfdataset(
                str(self.data_source),
                parallel=False,
                concat_dim=self.states_coord,
                combine="nested",
                data_vars="minimal",
                coords="minimal",
                compat="override",
            ) as ds:
                self.data_source = ds

        if sel is not None:
            self.data_source = self.data_source.sel(self.sel)
        if isel is not None:
            self.data_source = self.data_source.isel(self.isel)
        if pre_load:
            self.data_source.load()

        self._get_inds(self.data_source)

    def _get_inds(self, ds):
        """
        Helper function for index and weights
        reading
        """
        for c in [self.states_coord, self.x_coord, self.y_coord, self.h_coord]:
            if not c in ds:
                raise KeyError(
                    f"States '{self.name}': Missing coordinate '{c}' in data"
                )

        self._inds = ds[self.states_coord].to_numpy()
        if self.time_format is not None:
            self._inds = pd.to_datetime(self._inds, format=self.time_format).to_numpy()
        self._N = len(self._inds)

        if self.weight_ncvar is not None:
            self._weights = ds[self.weight_ncvar].to_numpy()

        for v in self.ovars:
            if v in self.var2ncvar:
                ncv = self.var2ncvar[v]
                if not ncv in ds:
                    raise KeyError(
                        f"States '{self.name}': nc variable '{ncv}' not found in data, found: {sorted(list(ds.keys()))}"
                    )
            elif v not in self.fixed_vars:
                raise ValueError(
                    f"States '{self.name}': Variable '{v}' neither found in var2ncvar not in fixed_vars"
                )

    def _get_data(self, ds, verbosity):
        """
        Helper function for data extraction
        """
        x = ds[self.x_coord].to_numpy()
        y = ds[self.y_coord].to_numpy()
        h = ds[self.h_coord].to_numpy()
        n_x = len(x)
        n_y = len(y)
        n_h = len(h)
        n_sts = ds.sizes[self.states_coord]

        cor_shxy = (self.states_coord, self.h_coord, self.x_coord, self.y_coord)
        cor_shyx = (self.states_coord, self.h_coord, self.y_coord, self.x_coord)
        cor_sh = (self.states_coord, self.h_coord)
        cor_s = (self.states_coord,)
        vars_shyx = []
        vars_sh = []
        vars_s = []
        for v, ncv in self.var2ncvar.items():
            if ds[ncv].dims == cor_shyx or ds[ncv].dims == cor_shxy:
                vars_shyx.append(v)
            elif ds[ncv].dims == cor_sh:
                vars_sh.append(v)
            elif ds[ncv].dims == cor_s:
                vars_s.append(v)
            else:
                raise ValueError(
                    f"States '{self.name}': Wrong coordinate order for variable '{ncv}': Found {ds[ncv].dims}, expecting {cor_shxy}, {cor_shyx}, {cor_sh} or {cor_s}"
                )

        data = np.zeros((n_sts, n_h, n_y, n_x, len(self.var2ncvar)), dtype=FC.DTYPE)
        for v in vars_shyx:
            ncv = self.var2ncvar[v]
            if ds[ncv].dims == cor_shyx:
                data[..., self._dkys[v]] = ds[ncv][:]
            else:
                data[..., self._dkys[v]] = np.swapaxes(ds[ncv].to_numpy(), 2, 3)
        for v in vars_sh:
            ncv = self.var2ncvar[v]
            data[..., self._dkys[v]] = ds[ncv].to_numpy()[:, :, None, None]
        for v in vars_s:
            ncv = self.var2ncvar[v]
            data[..., self._dkys[v]] = ds[ncv].to_numpy()[:, None, None, None]
        if FV.WD in self.fixed_vars:
            data[..., self._dkys[FV.WD]] = np.full(
                (n_sts, n_h, n_y, n_x), self.fixed_vars[FV.WD], dtype=FC.DTYPE
            )

        if verbosity > 1:
            print(f"\n{self.name}: Data ranges")
            for v, i in self._dkys.items():
                d = data[..., i]
                nn = np.sum(np.isnan(d))
                print(
                    f"  {v}: {np.nanmin(d)} --> {np.nanmax(d)}, nans: {nn} ({100*nn/len(d.flat):.2f}%)"
                )

        return data

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """

        if (FV.WS in self.ovars and FV.WD not in self.ovars) or (
            FV.WS not in self.ovars and FV.WD in self.ovars
        ):
            raise KeyError(
                f"States '{self.name}': Missing '{FV.WS}' or '{FV.WD}' in output variables {self.ovars}"
            )

        # ensure WD and WS get the first two slots of data:
        self._dkys = {}
        if FV.WS in self.ovars:
            self._dkys[FV.WD] = 0
        if FV.WS in self.var2ncvar:
            self._dkys[FV.WS] = 1
        for v in self.var2ncvar:
            if v not in self._dkys:
                self._dkys[v] = len(self._dkys)
        self._n_dvars = len(self._dkys)

        if self._weights is None:
            self._weights = np.full(
                (self._N, algo.n_turbines), 1.0 / self._N, dtype=FC.DTYPE
            )

        idata = super().load_data(algo, verbosity)

        if self.pre_load:
            self.X = self.var(FV.X)
            self.Y = self.var(FV.Y)
            self.H = self.var(FV.H)
            self.VARS = self.var("vars")
            self.DATA = self.var("data")

            ds = self.data_source

            h = ds[self.h_coord].to_numpy()
            y = ds[self.y_coord].to_numpy()
            x = ds[self.x_coord].to_numpy()
            v = list(self._dkys.keys())
            coos = (FC.STATE, self.H, self.Y, self.X, self.VARS)
            data = self._get_data(ds, verbosity)
            data = (coos, data)

            idata["coords"][self.H] = h
            idata["coords"][self.Y] = y
            idata["coords"][self.X] = x
            idata["coords"][self.VARS] = v
            idata["data_vars"][self.DATA] = data

        return idata

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def index(self):
        """
        The index list

        Returns
        -------
        indices: array_like
            The index labels of states, or None for default integers

        """
        return self._inds

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.ovars

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return self._weights

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        # prepare:
        n_states = tdata.n_states
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        points = tdata[FC.TARGETS].reshape(n_states, n_targets * n_tpoints, 3)
        n_pts = points.shape[1]
        n_states = fdata.n_states

        # pick pre-loaded data:
        if self.pre_load:
            x = mdata[self.X]
            y = mdata[self.Y]
            h = mdata[self.H]
            data = mdata[self.DATA].copy()

        # read data for this chunk:
        else:
            i0 = np.where(self._inds == mdata[FC.STATE][0])[0][0]
            s = slice(i0, i0 + n_states)
            ds = self.data_source.isel({self.states_coord: s}).load()

            x = ds[self.x_coord].to_numpy()
            y = ds[self.y_coord].to_numpy()
            h = ds[self.h_coord].to_numpy()
            data = self._get_data(ds, verbosity=0)

            del ds
        n_h = len(h)
        n_y = len(y)
        n_x = len(x)

        # translate WS, WD into U, V:
        if FV.WD in self.ovars and FV.WS in self.ovars:
            wd = data[..., self._dkys[FV.WD]]
            ws = (
                data[..., self._dkys[FV.WS]]
                if FV.WS in self._dkys
                else self.fixed_vars[FV.WS]
            )
            wdwsi = [self._dkys[FV.WD], self._dkys[FV.WS]]
            data[..., wdwsi] = wd2uv(wd, ws, axis=-1)
            del ws, wd

        # prepare points:
        sts = np.arange(n_states)
        pts = np.append(points, np.zeros((n_states, n_pts, 1), dtype=FC.DTYPE), axis=2)
        pts[:, :, 3] = sts[:, None]
        pts = pts.reshape(n_states * n_pts, 4)
        pts = np.flip(pts, axis=1)
        gvars = (sts, h, y, x)

        # interpolate nan values:
        if self.interp_nans and np.any(np.isnan(data)):
            df = pd.DataFrame(
                index=pd.MultiIndex.from_product(
                    gvars, names=["state", "height", "y", "x"]
                ),
                data={
                    v: data[..., vi].reshape(n_states * n_h * n_y * n_x)
                    for v, vi in self._dkys.items()
                },
            )
            df.interpolate(
                axis=0, method="linear", limit_direction="forward", inplace=True
            )
            df.interpolate(
                axis=0, method="linear", limit_direction="backward", inplace=True
            )
            data = df.to_numpy().reshape(n_states, n_h, n_y, n_x, self._n_dvars)
            del df

        # interpolate:
        try:
            data = interpn(gvars, data, pts, **self.interpn_pars).reshape(
                n_states, n_pts, self._n_dvars
            )
        except ValueError as e:
            print(f"\n\nStates '{self.name}': Interpolation error")
            print("INPUT VARS: (state, heights, y, x)")
            print(
                "DATA BOUNDS:", [np.min(d) for d in gvars], [np.max(d) for d in gvars]
            )
            print(
                "EVAL BOUNDS:", [np.min(p) for p in pts.T], [np.max(p) for p in pts.T]
            )
            raise e
        del pts, x, y, h, gvars

        # interpolate nan values:
        if self.interp_nans and np.any(np.isnan(data)):
            df = pd.DataFrame(
                index=pd.MultiIndex.from_product(
                    (sts, range(n_pts)), names=["state", "point"]
                ),
                data={
                    v: data[:, :, vi].reshape(n_states * n_pts)
                    for v, vi in self._dkys.items()
                },
            )
            df["x"] = points[:, :, 0].reshape(n_states * n_pts)
            df["y"] = points[:, :, 1].reshape(n_states * n_pts)
            df = df.reset_index().set_index(["state", "x", "y"])
            df.interpolate(
                axis=0, method="linear", limit_direction="forward", inplace=True
            )
            df.interpolate(
                axis=0, method="linear", limit_direction="backward", inplace=True
            )
            df = df.reset_index().drop(["x", "y"], axis=1).set_index(["state", "point"])
            data = df.to_numpy().reshape(n_states, n_pts, self._n_dvars)
            del df

        # set output:
        out = {}
        if FV.WD in self.ovars and FV.WS in self.ovars:
            uv = data[..., wdwsi]
            out[FV.WS] = np.linalg.norm(uv, axis=-1)
            out[FV.WD] = uv2wd(uv, axis=-1)
            del uv
        for v in self.ovars:
            if v not in out:
                if v in self._dkys:
                    out[v] = data[..., self._dkys[v]]
                else:
                    out[v] = np.full(
                        (n_states, n_pts), self.fixed_vars[v], dtype=FC.DTYPE
                    )

        return {v: d.reshape(n_states, n_targets, n_tpoints) for v, d in out.items()}



import numpy as np
from xarray import Dataset

from foxes.algorithms.downwind.downwind import Downwind
import foxes.constants as FC
import foxes.variables as FV
from foxes.core.data import MData, FData, TData

from . import models as mdls


class Sequential(Downwind):
    """
    A sequential calculation of states without chunking.

    This is of use for the evaluation in simulation
    environments that do not support multi-state computations,
    like FMUs.

    Attributes
    ----------
    ambient: bool
        Flag for ambient calculation
    calc_pars: dict
        Parameters for model calculation.
        Key: model name str, value: parameter dict
    states0: foxes.core.States
        The original states
    points: numpy.ndarray
        The points of interest, shape: (n_states, n_points, 3)
    plugins: list of foxes.algorithm.sequential.SequentialIterPlugin
        The plugins, updated with every iteration
    outputs: list of str
        The output variables
    :group: algorithms.sequential

    """

    @classmethod
    def get_model(cls, name):
        """
        Get the algorithm specific model

        Parameters
        ----------
        name: str
            The model name

        Returns
        -------
        model: foxes.core.model
            The model

        """
        try:
            return getattr(mdls, name)
        except AttributeError:
            return super().get_model(name)

    def __init__(
        self,
        farm,
        states,
        *args,
        points=None,
        ambient=False,
        calc_pars={},
        chunks={FC.STATE: None, FC.POINT: 4000},
        plugins=[],
        outputs=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        farm: foxes.WindFarm
            The wind farm
        states: foxes.core.States
            The ambient states
        args: tuple, optional
            Additional arguments for Downwind
        points: numpy.ndarray, optional
            The points of interest, shape: (n_states, n_points, 3)
        ambient: bool
            Flag for ambient calculation
        calc_pars: dict
            Parameters for model calculation.
            Key: model name str, value: parameter dict
        chunks: dict
            The xarray.Dataset chunk parameters
        plugins: list of foxes.algorithm.sequential.SequentialIterPlugin
            The plugins, updated with every iteration
        outputs: list of str, optional
            The output variables
        kwargs: dict, optional
            Additional arguments for Downwind

        """
        super().__init__(farm, mdls.SeqState(states), *args, chunks=chunks, **kwargs)
        self.ambient = ambient
        self.calc_pars = calc_pars
        self.states0 = self.states.states
        self.points = points
        self.plugins = plugins
        self.outputs = outputs if outputs is not None else self.DEFAULT_FARM_OUTPUTS

        self._i = None

    @property
    def iterating(self):
        """
        Flag for running iteration

        Returns
        -------
        itr: bool
            True if currently iterating

        """
        return self._i is not None

    def __iter__(self):
        """Initialize the iterator"""

        if not self.iterating:
            if not self.initialized:
                self.initialize()
            self._print_deco("calc_farm")

            self._inds = self.states0.index()
            self._weights = self.states0.weights(self)
            self._i = 0
            self._counter = 0

            self._mlist, self._calc_pars = self._collect_farm_models(
                self.outputs, self.calc_pars, self.ambient
            )
            if not self._mlist.initialized:
                self._mlist.initialize(self, self.verbosity)
                self._calc_farm_vars(self._mlist)
            self._print_model_oder(self._mlist, self._calc_pars)

            self._mdata = self.get_models_idata()
            if self.verbosity > 0:
                s = "\n".join(
                    [
                        f"  {v}: {d[0]} {d[1].dtype}, shape {d[1].shape}"
                        for v, d in self._mdata["data_vars"].items()
                    ]
                )
                print("\nInput data:\n")
                print(s, "\n")
                print(f"Output farm variables:", ", ".join(self.farm_vars))
                print()

            self._mdata = MData(
                data={v: d[1] for v, d in self._mdata["data_vars"].items()},
                dims={v: d[0] for v, d in self._mdata["data_vars"].items()},
                loop_dims=[FC.STATE],
                name="mdata",
            )

            self._fdata = FData(
                data={
                    v: np.zeros((self.n_states, self.n_turbines), dtype=FC.DTYPE)
                    for v in self.farm_vars
                },
                dims={v: (FC.STATE, FC.TURBINE) for v in self.farm_vars},
                loop_dims=[FC.STATE],
                name="fdata",
            )

            if self.points is not None:
                self._plist, self._calc_pars_p = self._collect_point_models(
                    ambient=self.ambient
                )
                if not self._plist.initialized:
                    self._plist.initialize(self, self.verbosity)
                self._pvars = self._plist.output_point_vars(self)
                self.print(f"\nOutput point variables:", ", ".join(self._pvars), "\n")

                n_points = self.points.shape[1]
                self._tdata = TData.from_points(
                    self.points,
                    data={
                        v: np.zeros((self.n_states, n_points, 1), dtype=FC.DTYPE)
                        for v in self._pvars
                    },
                    dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in self._pvars},
                )

            for p in self.plugins:
                p.initialize(self)

        return self

    def __next__(self):
        """Run calculation for current step, then iterate to next"""

        if self._i < len(self._inds):
            self._counter = self._i
            self.states._counter = self._i
            self.states._size = 1
            self.states._indx = self._inds[self._i]
            self.states._weight = self._weights[self._i]

            mdata = MData(
                data={
                    v: d[self._i, None] if self._mdata.dims[v][0] == FC.STATE else d
                    for v, d in self._mdata.items()
                },
                dims={v: d for v, d in self._mdata.dims.items()},
                loop_dims=[FC.STATE],
                name="mdata",
            )

            fdata = FData(
                data={
                    v: np.zeros((1, self.n_turbines), dtype=FC.DTYPE)
                    for v in self.farm_vars
                },
                dims={v: (FC.STATE, FC.TURBINE) for v in self.farm_vars},
                loop_dims=[FC.STATE],
                name="fdata",
            )

            fres = self._mlist.calculate(self, mdata, fdata, parameters=self._calc_pars)
            fres[FV.WEIGHT] = self.weight[None, :]

            for v, d in fres.items():
                self._fdata[v][self._i] = d[0]

            fres = Dataset(
                coords={FC.STATE: [self.index]},
                data_vars={v: ((FC.STATE, FC.TURBINE), d) for v, d in fres.items()},
            )
            fres[FC.TNAME] = ((FC.TURBINE,), self.farm.turbine_names)
            if FV.ORDER in fres:
                fres[FV.ORDER] = fres[FV.ORDER].astype(FC.ITYPE)

            if self.points is None:
                for p in self.plugins:
                    p.update(self, fres)

                self._i += 1
                return fres

            else:
                n_points = self.points.shape[1]
                tdata = TData.from_points(
                    self.points[self.counter, None],
                    data={
                        v: np.zeros((1, n_points, 1), dtype=FC.DTYPE)
                        for v in self._pvars
                    },
                    dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in self._pvars},
                )

                pres = self._plist.calculate(
                    self, mdata, fdata, tdata, parameters=self._calc_pars_p
                )

                for v, d in pres.items():
                    self._tdata[v][self._i] = d[0]

                pres = Dataset(
                    coords={FC.STATE: [self.index]},
                    data_vars={
                        v: ((FC.STATE, FC.TARGET, FC.TPOINT), d)
                        for v, d in pres.items()
                    },
                )

                for p in self.plugins:
                    p.update(self, fres, pres)

                self._i += 1
                return fres, pres

        else:
            del self._mdata

            self._i = None
            self.states._counter = None
            self.states._size = len(self._inds)
            self.states._indx = self._inds
            self.states._weight = self._weights

            for p in self.plugins:
                p.finalize(self)

            raise StopIteration

    @property
    def size(self):
        """
        The total number of iteration steps

        Returns
        -------
        s: int
            The total number of iteration steps

        """
        return self.states.size()

    @property
    def counter(self):
        """
        The current index counter

        Returns
        -------
        i: int
            The current index counter

        """
        return self._counter if self.iterating else None

    @property
    def index(self):
        """
        The current index

        Returns
        -------
        indx: int
            The current index

        """
        return self.states._indx if self.iterating else None

    def states_i0(self, counter, algo=None):
        """
        Returns counter or index

        Parameters
        ----------
        counter: bool
            Flag for counter
        algo: object, optional
            Dummy argument, due to consistency with
            foxes.core.Data.states_i0

        Returns
        -------
        i0: int
            The counter or index

        """
        return self.counter if counter else self.index

    @property
    def weight(self):
        """
        The current weight array

        Returns
        -------
        w: numpy.ndarray
            The current weight array, shape: (n_turbines,)

        """
        return self.states._weight if self.iterating else None

    @property
    def mdata(self):
        """
        Get the current model data

        Returns
        -------
        d: foxes.core.MData
            The current model data

        """
        return self._mdata if self.iterating else None

    @property
    def fdata(self):
        """
        Get the current farm data

        Returns
        -------
        d: foxes.core.FData
            The current farm data

        """
        return self._fdata

    @property
    def tdata(self):
        """
        Get the current point data

        Returns
        -------
        d: foxes.core.TData
            The current point data

        """
        return self._tdata if self.points is not None and self.iterating else None

    @property
    def farm_results(self):
        """
        The overall farm results

        Returns
        -------
        results: xarray.Dataset
            The overall farm results

        """
        results = Dataset(
            coords={FC.STATE: self._inds, FC.TURBINE: np.arange(self.n_turbines)},
            data_vars={v: (self._fdata.dims[v], d) for v, d in self._fdata.items()},
        )

        results[FC.TNAME] = ((FC.TURBINE,), self.farm.turbine_names)
        if FV.ORDER in results:
            results[FV.ORDER] = results[FV.ORDER].astype(FC.ITYPE)

        return results

    @property
    def prev_farm_results(self):
        """
        Alias for farm_results

        Returns
        -------
        results: xarray.Dataset
            The overall farm results

        """
        return self.farm_results

    @property
    def cur_farm_results(self):
        """
        The current farm results

        Returns
        -------
        results: xarray.Dataset
            The current farm results

        """

        i = self.counter
        results = Dataset(
            coords={FC.STATE: [self.index], FC.TURBINE: np.arange(self.n_turbines)},
            data_vars={
                v: (self._fdata.dims[v], d[i, None]) for v, d in self._fdata.items()
            },
        )

        results[FC.TNAME] = ((FC.TURBINE,), self.farm.turbine_names)
        if FV.ORDER in results:
            results[FV.ORDER] = results[FV.ORDER].astype(FC.ITYPE)

        return results

    @property
    def point_results(self):
        """
        The overall point results

        Returns
        -------
        results: xarray.Dataset
            The overall point results

        """
        n_points = self.points.shape[1]
        results = Dataset(
            coords={
                FC.STATE: self._inds,
                FC.TURBINE: np.arange(self.n_turbines),
                FC.POINT: np.arange(n_points),
                FC.XYH: np.arange(3),
            },
            data_vars={v: (self._tdata.dims[v], d) for v, d in self._tdata.items()},
        )

        return results

    @property
    def cur_point_results(self):
        """
        The current point results

        Returns
        -------
        results: xarray.Dataset
            The current point results

        """

        n_points = self.points.shape[1]
        i = self.counter

        results = Dataset(
            coords={
                FC.STATE: [self.index],
                FC.TURBINE: np.arange(self.n_turbines),
                FC.POINT: np.arange(n_points),
                FC.XYH: np.arange(3),
            },
            data_vars={
                v: (self._tdata.dims[v], d[i, None]) for v, d in self._tdata.items()
            },
        )

        return results

    def calc_farm(self, *args, **kwargs):
        if not self.iterating:
            raise ValueError(f"calc_farm call is only allowed during iterations")

        return self.cur_farm_results

    def calc_points(self, farm_results, points):
        if not self.iterating:
            raise ValueError(f"calc_points call is only allowed during iterations")

        n_points = points.shape[1]

        plist, calc_pars = self._collect_point_models(ambient=self.ambient)
        if not plist.initialized:
            plist.initialize(self, self.verbosity)
        pvars = plist.output_point_vars(self)

        mdata = self.get_models_idata()
        mdata = MData(
            data={v: d[1] for v, d in mdata["data_vars"].items()},
            dims={v: d[0] for v, d in mdata["data_vars"].items()},
            loop_dims=[FC.STATE],
            name="mdata",
        )
        mdata = MData(
            data={
                v: d[self.states.counter, None] if mdata.dims[v][0] == FC.STATE else d
                for v, d in mdata.items()
            },
            dims={v: d for v, d in mdata.dims.items()},
            loop_dims=[FC.STATE],
            name="mdata",
        )

        fdata = FData(
            data={v: farm_results[v].to_numpy() for v in self.farm_vars},
            dims={v: (FC.STATE, FC.TURBINE) for v in self.farm_vars},
            loop_dims=[FC.STATE],
            name="fdata",
        )

        tdata = TData.from_points(
            points[0, None],
            data={v: np.zeros((1, n_points, 1), dtype=FC.DTYPE) for v in pvars},
            dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in pvars},
            name="tdata",
        )

        pres = plist.calculate(self, mdata, fdata, tdata, parameters=calc_pars)
        pres = Dataset(
            coords={FC.STATE: self.states.index()},
            data_vars={
                v: ((FC.STATE, FC.TARGET, FC.TPOINT), d) for v, d in pres.items()
            },
        )

        # plist.finalize(self, self.verbosity)

        return pres



from abc import abstractmethod

from .model import Model
from foxes.utils import all_subclasses


class VerticalProfile(Model):
    """
    Abstract base class for vertical profiles.

    :group: core

    """

    @abstractmethod
    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        return []

    @abstractmethod
    def calculate(self, tdata, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        tdata: dict
            The target point data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """
        pass

    @classmethod
    def new(cls, profile_type, **kwargs):
        """
        Run-time profile factory.

        Parameters
        ----------
        profile_type: str
            The selected derived class name

        """

        if profile_type is None:
            return None

        allc = all_subclasses(cls)
        found = profile_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == profile_type:
                    return scls(**kwargs)

        else:
            estr = "Vertical profile type '{}' is not defined, available types are \n {}".format(
                profile_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)



import foxes.variables as FV
from foxes.core import FarmDataModel


class SetAmbFarmResults(FarmDataModel):
    """
    This model copies farm data results to ambient results.

    Attributes
    ----------
    vars: list of str
        The variables to be copied, or `None` for all

    :group: algorithms.downwind.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__()
        self.vars = None

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        if self.vars is None:
            self.vars = set([v for v in algo.farm_vars if v in FV.var2amb])
        return [FV.var2amb[v] for v in self.vars]

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        for v in self.vars:
            fdata[FV.var2amb[v]] = fdata[v].copy()
        return {v: fdata[v] for v in self.output_farm_vars(algo)}



import matplotlib.animation as animation


class Animator:
    """
    Creates an animation from generators
    that yield lists of artists.

    Attributes
    ----------
    fig: pyplot.Figure
        The figure object

    :group: output

    """

    def __init__(self, fig=None):
        """
        Constructor.

        Parameters
        ----------
        fig: pyplot.Figure, optional
            The figure object

        """
        self.fig = fig
        self._gens = []

    def add_generator(self, gen):
        """
        Add a generator.

        Parameters
        ----------
        gen: Generator
            A generator that yields (fig, list of Artist)

        """
        self._gens.append(gen)

    @property
    def generators(self):
        """
        The artist generators

        Returns
        -------
        gens: list of generators
            Generators that yield (fig, list of Artist)

        """
        return self._gens

    def animate(self, verbosity=1, **kwargs):
        """
        Create the animation

        Parameters
        ----------
        verbostiy: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Arguments for pyplot.animation.ArtistAnimation

        Returns
        -------
        ani: pyplot.animation.ArtistAnimation
            The animation

        """
        if len(self.generators) == 0:
            return None

        if verbosity > 0:
            print("Creating animation data")

        si = 0
        arts = []
        while True:
            if verbosity > 1:
                print(f"  Frame {si}")

            harts = []
            for g in self.generators:
                try:
                    y = next(g)

                    if len(y) != 2:
                        raise ValueError(
                            f"Expecting yield (fig, artists) from generator {g}"
                        )

                    fig, artists = y
                    if self.fig is None:
                        self.fig = fig
                    elif fig is not self.fig:
                        raise ValueError(f"Wrong figure returned by generator {g}")

                    harts += [a for a in artists]

                except StopIteration:
                    pass

            if len(harts):
                arts.append(harts)
                si += 1
            else:
                break

        if verbosity > 1:
            print("Done.")

        kwa = dict(interval=200, blit=True, repeat_delay=2000)
        kwa.update(kwargs)
        ani = animation.ArtistAnimation(fig, arts, **kwa)

        return ani



import numpy as np
import pandas as pd
from pathlib import Path

from foxes.core import TurbineType
from foxes.utils import PandasFileHelper
from foxes.data import PCTCURVE, parse_Pct_file_name
import foxes.variables as FV
import foxes.constants as FC


class PCtFile(TurbineType):
    """
    Calculate power and ct by interpolating
    from power-ct-curve data file (or pandas DataFrame).

    Attributes
    ----------
    source: str or pandas.DataFrame
        The file path, static name, or data
    col_ws: str
        The wind speed column
    col_P: str
        The power column
    col_ct: str
        The ct column
    rho: float
        The air densitiy for which the data is valid
        or None for no correction
    WSCT: str
        The wind speed variable for ct lookup
    WSP: str
        The wind speed variable for power lookup
    rpars: dict, optional
        Parameters for pandas file reading

    :group: models.turbine_types

    """

    def __init__(
        self,
        data_source,
        col_ws="ws",
        col_P="P",
        col_ct="ct",
        rho=None,
        p_ct=1.0,
        p_P=1.88,
        var_ws_ct=FV.REWS2,
        var_ws_P=FV.REWS3,
        pd_file_read_pars={},
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            The file path, static name, or data
        col_ws: str
            The wind speed column
        col_P: str
            The power column
        col_ct: str
            The ct column
        rho: float, optional
            The air densitiy for which the data is valid
            or None for no correction
        p_ct: float
            The exponent for yaw dependency of ct
        p_P: float
            The exponent for yaw dependency of P
        var_ws_ct: str
            The wind speed variable for ct lookup
        var_ws_P: str
            The wind speed variable for power lookup
        pd_file_read_pars: dict
            Parameters for pandas file reading
        parameters: dict, optional
            Additional parameters for TurbineType class

        """
        if not isinstance(data_source, pd.DataFrame):
            pars = parse_Pct_file_name(data_source)
            pars.update(parameters)
        else:
            pars = parameters

        super().__init__(**pars)

        self.source = data_source
        self.col_ws = col_ws
        self.col_P = col_P
        self.col_ct = col_ct
        self.rho = rho
        self.p_ct = p_ct
        self.p_P = p_P
        self.WSCT = var_ws_ct
        self.WSP = var_ws_P
        self.rpars = pd_file_read_pars

    def __repr__(self):
        a = f"D={self.D}, H={self.H}, P_nominal={self.P_nominal}, P_unit={self.P_unit}, rho={self.rho}"
        a += f", var_ws_ct={self.WSCT}, var_ws_P={self.WSP}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.P, FV.CT]

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        if isinstance(self.source, pd.DataFrame):
            data = self.source
        else:
            fpath = algo.dbook.get_file_path(PCTCURVE, self.source, check_raw=True)
            if verbosity > 0:
                if not Path(self.source).is_file():
                    print(
                        f"Turbine type '{self.name}': Reading static data from context '{PCTCURVE}'"
                    )
                    print(f"Path: {fpath}")
                else:
                    print(f"Turbine type '{self.name}': Reading file", self.source)
            data = PandasFileHelper.read_file(fpath, **self.rpars)

        data = data.set_index(self.col_ws).sort_index()
        self.data_ws = data.index.to_numpy()
        self.data_P = data[self.col_P].to_numpy()
        self.data_ct = data[self.col_ct].to_numpy()

        if self.P_nominal is None:
            self.P_nominal = np.max(self.data_P) / FC.P_UNITS[self.P_unit]
            if verbosity > 0:
                print(
                    f"Turbine type '{self.name}': Setting P_nominal = {self.P_nominal:.2f} {self.P_unit}"
                )

        return super().load_data(algo, verbosity)

    def modify_cutin(
        self,
        modify_ct,
        modify_P,
        steps=20,
        iterations=100,
        a=0.55,
        b=0.55,
    ):
        """
        Modify the data such that a discontinuity
        at cutin wind speed is avoided

        Parameters
        ----------
        variable: str
            The target variable
        modify_ct: bool
            Flag for modification of the ct curve
        modify_P: bool
            Flag for modification of the power curve
        steps: int
            The number of wind speed steps between 0 and
            the cutin wind speed
        iterations: int
            The number of iterations
        a: float
            Coefficient for iterative mixing
        b: float
            Coefficient for iterative mixing

        """
        if modify_ct or modify_P:

            ws = self.data_ws
            ct = self.data_ct
            P = self.data_P

            i = 0
            try:
                while (
                    i < len(ws)
                    and (not modify_ct or ct[i] < 1e-5)
                    and (not modify_P or P[i] < 0.1)
                ):
                    i += 1
            except IndexError:
                raise IndexError(
                    f"Turbine type '{self.name}': Failed not determine cutin wind speed. ws = {ws}, ct = {ct}, P = {P}"
                )

            if ws[i] > 0:
                ws = ws[i:]
                ct = ct[i:]
                P = P[i:]

                new_ws = np.linspace(0.0, ws[0], steps + 1, dtype=ws.dtype)
                new_ct = np.zeros_like(new_ws)
                new_P = np.zeros_like(new_ws)

                if modify_ct:
                    new_ct[-1] = ct[0]
                    for it in range(iterations):
                        new_ct[1:-1] = a * new_ct[:-2] + (1 - a) * new_ct[2:]

                if modify_P:
                    new_P[-1] = P[0]
                    for it in range(iterations):
                        new_P[1:-1] = b * new_P[:-2] + (1 - b) * new_P[2:]

                self.data_ws = np.concatenate([new_ws[:-1], ws], axis=0)
                self.data_ct = np.concatenate([new_ct[:-1], ct], axis=0)
                self.data_P = np.concatenate([new_P[:-1], P], axis=0)

        else:
            super().modify_cutin(modify_ct, modify_P)

    def calculate(self, algo, mdata, fdata, st_sel):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: numpy.ndarray of bool
            The state-turbine selection,
            shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        rews2 = fdata[self.WSCT][st_sel]
        rews3 = fdata[self.WSP][st_sel]

        # apply air density correction:
        if self.rho is not None:
            # correct wind speed by air density, such
            # that in the partial load region the
            # correct value is reconstructed:
            rho = fdata[FV.RHO][st_sel]
            # rews2 *= (self.rho / rho) ** 0.5
            rews3 *= (self.rho / rho) ** (1.0 / 3.0)
            del rho

        # in yawed case, calc yaw corrected wind speed:
        if FV.YAWM in fdata and (self.p_P is not None or self.p_ct is not None):
            # calculate corrected wind speed wsc,
            # gives ws**3 * cos**p_P in partial load region
            # and smoothly deals with full load region:
            yawm = fdata[FV.YAWM][st_sel]
            if np.any(np.isnan(yawm)):
                raise ValueError(
                    f"{self.name}: Found NaN values for variable '{FV.YAWM}'. Maybe change order in turbine_models?"
                )
            cosm = np.cos(yawm / 180 * np.pi)
            if self.p_ct is not None:
                rews2 *= (cosm**self.p_ct) ** 0.5
            if self.p_P is not None:
                rews3 *= (cosm**self.p_P) ** (1.0 / 3.0)
            del yawm, cosm

        out = {
            FV.P: fdata[FV.P],
            FV.CT: fdata[FV.CT],
        }

        out[FV.P][st_sel] = np.interp(
            rews3, self.data_ws, self.data_P, left=0.0, right=0.0
        )
        out[FV.CT][st_sel] = np.interp(
            rews2, self.data_ws, self.data_ct, left=0.0, right=0.0
        )

        return out

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level

        """
        super().finalize(algo, verbosity)
        del self.data_ws, self.data_P, self.data_ct



from foxes.core import TurbineModel
import foxes.variables as FV
import foxes.constants as FC
from foxes.utils import delta_wd


class YAW2YAWM(TurbineModel):
    """
    Calculates delta yaw (i.e. YAWM) from absolute
    yaw (i.e. YAW)

    :group: models.turbine_models

    """

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.YAWM]

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        yaw = fdata[FV.YAW][st_sel]
        wd = fdata[FV.WD][st_sel]

        yawm = fdata[FV.YAWM]
        yawm[st_sel] = delta_wd(wd, yaw)

        return {FV.YAWM: yawm}



"""
Functions for usingin windIO yaml files as input.
"""

from .windio import read_windio
from .read_fields import wio2foxes, foxes2wio
from .get_states import get_states
from .read_farm import read_turbine_type, read_layout
from .read_attributes import read_attributes
from .read_outputs import read_outputs
from .runner import WindioRunner



from foxes.core import FarmDataModel


class URelax(FarmDataModel):
    """
    Apply under-relaxation to selected variables
    during iterations.

    Attributes
    ----------
    urel: dict
        The variables and their under-relaxation
        factors bewteen 0 and 1

    :group: algorithms.iterative.models

    """

    def __init__(self, **urel):
        """
        Constructor.

        Parameters
        ----------
        urel: dict
            The variables and their under-relaxation
            factors bewteen 0 and 1

        """
        super().__init__()
        self.urel = urel
        self.name += "_" + "_".join(list(urel.keys()))

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return list(self.urel.keys())

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        i0 = fdata.states_i0(counter=True, algo=algo)
        i1 = i0 + fdata.n_states
        pres = algo.prev_farm_results

        out = {}
        for v, u in self.urel.items():
            if u > 0 and pres is not None:
                odata = pres[v].to_numpy()[i0:i1]
                out[v] = u * odata + (1 - u) * fdata[v]
            else:
                out[v] = fdata[v]

        return out



import numpy as np

from foxes.core import Turbine


def add_row(
    farm,
    xy_base,
    xy_step,
    n_turbines,
    indices=None,
    names=None,
    verbosity=1,
    **turbine_parameters
):
    """
    Add a single row of turbines.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    xy_base: numpy.ndarray
        The base point, shape: (2,)
    xy_step: numpy.ndarray
        The step vector, shape: (2,)
    n_turbines: int
        The number of turbines
    indices: list of int, optional
        The turbine indices
    names: list of str, optional
        The turbine names
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Parameters forwarded to `foxes.core.Turbine`

    :group: input.farm_layout

    """
    p0 = np.array(xy_base)
    delta = np.array(xy_step)

    for i in range(n_turbines):
        farm.add_turbine(
            Turbine(
                xy=p0 + i * delta,
                index=None if indices is None else indices[i],
                name=None if names is None else names[i],
                **turbine_parameters
            ),
            verbosity=verbosity,
        )



from abc import abstractmethod
import numpy as np

from foxes.utils import all_subclasses, Factory
import foxes.variables as FV
import foxes.constants as FC

from .model import Model


class WakeModel(Model):
    """
    Abstract base class for wake models.

    :group: core

    """

    @property
    def affects_downwind(self):
        """
        Flag for downwind or upwind effects
        on other turbines

        Returns
        -------
        dwnd: bool
            Flag for downwind effects by this model

        """
        return True

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_targets, n_tpoints, ...)

        """
        return {FV.WS: np.zeros_like(tdata[FC.TARGETS][..., 0])}

    @abstractmethod
    def contribute(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        wake_coos,
        wake_deltas,
    ):
        """
        Modifies wake deltas at target points by
        contributions from the specified wake source turbines.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints, ...)

        """
        pass

    def finalize_wake_deltas(
        self,
        algo,
        mdata,
        fdata,
        amb_results,
        wake_deltas,
    ):
        """
        Finalize the wake calculation.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        amb_results: dict
            The ambient results, key: variable name str,
            values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)
        wake_deltas: dict
            The wake deltas object at the selected target
            turbines. Key: variable str, value: numpy.ndarray
            with shape (n_states, n_targets, n_tpoints)

        """
        pass

    @classmethod
    def new(cls, wmodel_type, *args, **kwargs):
        """
        Run-time wake model factory.

        Parameters
        ----------
        wmodel_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if wmodel_type is None:
            return None

        allc = all_subclasses(cls)
        found = wmodel_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == wmodel_type:
                    return scls(*args, **kwargs)

        else:
            estr = (
                "Wake model type '{}' is not defined, available types are \n {}".format(
                    wmodel_type, sorted([i.__name__ for i in allc])
                )
            )
            raise KeyError(estr)


class TurbineInductionModel(WakeModel):
    """
    Abstract base class for turbine induction models.

    :group: core

    """

    @property
    def affects_downwind(self):
        """
        Flag for downwind or upwind effects
        on other turbines

        Returns
        -------
        dwnd: bool
            Flag for downwind effects by this model

        """
        return False


class WakeK(Model):
    """
    Handler for the wake growth parameter k

    Attributes
    ----------
    k_var: str
        The name of the k variable
    ti_var: str
        The name of the TI variable

    :group: core

    """

    def __init__(
        self,
        k=None,
        ka=None,
        kb=None,
        k_var=FV.K,
        ti_var=FV.TI,
    ):
        """
        Constructor.

        Parameters
        ----------
        k: float, optional
            The k value
        ka: float, optional
            The ka value in k = ka * TI + kb
        kb: float, optional
            The kb value in k = ka * TI + kb
        k_var: str
            The name of the k variable
        ti_var: str
            The name of the TI variable

        """
        super().__init__()
        self._k = k
        self._ka = ka
        self._kb = kb
        self.k_var = k_var
        self.ti_var = ti_var

        if k is not None and (ka is not None or kb is not None):
            raise ValueError("Got 'k' and also ('ka' or 'kb') as non-None parameters")
        elif k is None and kb is not None and (ka is None or ka == 0):
            raise ValueError(f"Got k={k}, ka={ka}, kb={kb}, use k={kb} instead")

        setattr(self, self.k_var, None)

    def repr(self):
        """
        Provides the representative string

        Returns
        -------
        s: str
            The representative string

        """
        if self._k is not None:
            s = f"{self.k_var}={self._k}"
        elif self._ka is not None or self._kb is not None:
            s = f"{self.k_var}={self._ka}*{self.ti_var}"
            if self._kb is not None and self._kb > 0:
                s += f"+{self._kb}"
        else:
            s = f"k_var={self.k_var}"
        return s

    @property
    def all_none(self):
        """Flag for k=ka=kb=None"""
        return self._k is None and self._ka is None and self._kb is None

    @property
    def use_amb_ti(self):
        """Flag for using ambient ti"""
        return self.ti_var in FV.amb2var

    def __call__(
        self,
        *args,
        lookup_ti="w",
        lookup_k="sw",
        ti=None,
        amb_ti=None,
        **kwargs,
    ):
        """
        Gets the k value

        Parameters
        ----------
        args: tuple, optional
            Arguments for get_data
        lookup_ti: str
            The ti lookup order for get_data
        lookup_k: str
            The k lookup order for get_data
        ti: numpy.ndarray, optional
            ti data in the requested target shape,
            if known
        amb_ti: numpy.ndarray, optional
            Ambient ti data in the requested target shape,
            if known
        kwargs: dict, optional
            Arguments for get_data

        Returns
        -------
        k: numpy.ndarray
            The k array as returned by get_data

        """
        if self._k is not None:
            setattr(self, self.k_var, self._k)
        elif self._ka is not None or self._kb is not None:
            if self.ti_var == FV.TI and ti is not None:
                pass
            elif self.ti_var == FV.AMB_TI and amb_ti is not None:
                ti = amb_ti
            else:
                ti = self.get_data(self.ti_var, *args, lookup=lookup_ti, **kwargs)
            kb = 0 if self._kb is None else self._kb
            setattr(self, self.k_var, self._ka * ti + kb)

        k = self.get_data(self.k_var, *args, lookup=lookup_k, **kwargs)
        setattr(self, self.k_var, None)
        return k



import numpy as np
import matplotlib.pyplot as plt

from .area_geometry import AreaGeometry


class Circle(AreaGeometry):
    """
    This class represents the area of a circle.

    Attributes
    ----------
    centre: numpy.ndarray
        The centre point, shape: (2,)
    radius: float
        The radius

    :group: utils.geom2d

    """

    def __init__(self, centre, radius):
        """
        Cobnstructor.

        Parameters
        ----------
        centre: numpy.ndarray
            The centre point, shape: (2,)
        radius: float
            The radius

        """
        self.centre = np.array(centre, dtype=np.float64)
        self.radius = radius

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        return self.centre - self.radius

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        return self.centre + self.radius

    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """

        deltas = points - self.centre[None, :]
        magd = np.linalg.norm(deltas, axis=-1)
        dists = np.abs(magd - self.radius)

        if return_nearest:
            sel = magd > 0.0
            if np.all(sel):
                minp = self.centre + deltas / magd[:, None] * self.radius
            else:
                minp = np.zeros_like(points)
                minp[sel] = deltas[sel] / magd[sel]
                minp[~sel][:, 0] = 1
                minp = self.centre + minp * self.radius
            return dists, minp
        else:
            return dists

    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        magd = np.linalg.norm(points - self.centre[None, :], axis=-1)
        return magd <= self.radius

    def add_to_figure(
        self, ax, show_boundary=True, fill_mode=None, pars_boundary={}, pars_distance={}
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        if show_boundary:
            pars = dict(color="darkblue", linewidth=1, fill=False)
            pars.update(pars_boundary)

            circle = plt.Circle(self.centre, self.radius, **pars)
            ax.add_patch(circle)

        super().add_to_figure(
            ax, show_boundary, fill_mode, pars_boundary, pars_distance
        )


if __name__ == "__main__":
    centre = np.array([3.0, 4.0])
    radius = 2.5
    N = 500

    fig, ax = plt.subplots()
    g = Circle(centre, radius)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = Circle(centre, radius)
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = Circle(centre, radius).inverse()
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = Circle(centre, radius).inverse()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)



from foxes.utils import Dict
from foxes.core import WakeModel, WakeFrame
import foxes.variables as FV

from .read_outputs import read_outputs


def _read_wind_deficit(wind_deficit, superposition, induction, algo_dict, verbosity):
    """Reads the wind deficit wake model"""

    wind_def_map = Dict(
        {
            "Jensen": "JensenWake",
            "Bastankhah2014": "Bastankhah2014",
            "Bastankhah2016": "Bastankhah2016",
            "TurbOPark": "TurbOPark",
        },
        name="wind_def_map",
    )

    ws_sup_dict = Dict(
        {
            "Linear": "ws_linear",
            "Quadratic": "ws_quadratic",
        },
        name="ws_sup_dict",
    )
    ws_sup_amb_dict = Dict(
        {
            "Linear": "ws_linear_amb",
            "Quadratic": "ws_quadratic_amb",
        },
        name="ws_sup_dict",
    )
    
    wname = wind_deficit.pop("name")
    eff_ws = wind_deficit.pop("use_effective_ws", True)
    if verbosity > 2:
        print("    Reading wind_deficit_model")
        print("      Name    :", wname)
        print("      Eff ws  :", eff_ws)
        print("      Contents:", [k for k in wind_deficit.keys()])
    wind_def_dict = Dict(wmodel_type=wind_def_map[wname], induction=induction)
    kcoef = Dict(wind_deficit["wake_expansion_coefficient"], name="kcoef")
    ka = kcoef["k_a"]
    kb = kcoef.get("k_b", 0.0)
    amb_ti = kcoef.get("free_stream_ti", False)
    if ka is None or ka == 0.0:
        wind_def_dict["k"] = kb
        if verbosity > 2:
            print("        Using k =", kb)
    else:
        ti_var = FV.AMB_TI if amb_ti else FV.TI
        if verbosity > 2:
            print(f"      Using k = {ka} * {ti_var} + {kb}")
        wind_def_dict["k"] = None
        wind_def_dict["ka"] = ka
        wind_def_dict["kb"] = kb
        wind_def_dict["ti_var"] = ti_var
    if "ceps" in wind_deficit:
        sbf = wind_deficit["ceps"]
        if verbosity > 2:
            print(f"      Using sbeta_factor = {sbf}")
        wind_def_dict["sbeta_factor"] = sbf
    supd = ws_sup_dict if eff_ws else ws_sup_amb_dict
    wind_def_dict["superposition"] = supd[superposition["ws_superposition"]]

    algo_dict["mbook"].wake_models[wname] = WakeModel.new(**wind_def_dict)
    if verbosity > 2:
        print(f"      Created wake model '{wname}':")
        print("       ", algo_dict["mbook"].wake_models[wname])
    algo_dict["wake_models"].append(wname)

    return ka, kb, amb_ti

def _read_turbulence(
    turbulence_model, superposition, induction, algo_dict, ka, kb, amb_ti, verbosity
):
    """Reads the ti wake model"""

    twake_def_map = Dict(
        {
            "CrespoHernandez": "CrespoHernandezTIWake",
            "IEC-TI-2019": "IECTIWake",
        },
        name="twake_def_map",
    )

    ti_sup_dict = Dict(
        {
            "Linear": "ti_linear",
            "Quadratic": "ti_quadratic",
        },
        name="ti_sup_dict",
    )

    wname = turbulence_model.pop("name")
    if verbosity > 2:
        print("    Reading turbulence_model")
        print("      Name:", wname)
        print("      Contents:", [k for k in turbulence_model.keys()])
    if wname != "None":
        tiwake_dict = dict(wmodel_type=twake_def_map[wname], induction=induction)
        if wname == "IEC-TI-2019":
            tiwake_dict["opening_angle"] = None
            tiwake_dict["iec_type"] = "2019"
        if "wake_expansion_coefficient" in turbulence_model:
            kcoef = Dict(turbulence_model["wake_expansion_coefficient"], name="kcoef")
            ka = kcoef["k_a"]
            kb = kcoef.get("k_b", 0.0)
            amb_ti = kcoef.get("free_stream_ti", False)
        if ka is None or ka == 0.0:
            tiwake_dict["k"] = kb
            if verbosity > 2:
                print("        Using k =", kb)
        else:
            ti_var = FV.AMB_TI if amb_ti else FV.TI
            if verbosity > 2:
                print(f"      Using k = {ka} * {ti_var} + {kb}")
            tiwake_dict["k"] = None
            tiwake_dict["ka"] = ka
            tiwake_dict["kb"] = kb
            tiwake_dict["ti_var"] = ti_var
        tiwake_dict["superposition"] = ti_sup_dict[superposition["ti_superposition"]]

        algo_dict["mbook"].wake_models[wname] = WakeModel.new(**tiwake_dict)
        if verbosity > 2:
            print(f"      Created wake model '{wname}':")
            print("       ", algo_dict["mbook"].wake_models[wname])
        algo_dict["wake_models"].append(wname)

def _read_blockage(blockage_model, induction, algo_dict, verbosity):
    """Reads the blockage model"""
    indc_def_map = Dict(
        {
            "RankineHalfBody": "RankineHalfBody",
            "Rathmann": "Rathmann",
            "SelfSimilarityDeficit": "SelfSimilar",
            "SelfSimilarityDeficit2020": "SelfSimilar2020",
        },
        name="twake_def_map",
    )

    wname = blockage_model.pop("name")
    if verbosity > 2:
        print("    Reading blockage_model")
        print("      Name:", wname)
        print("      Contents:", [k for k in blockage_model.keys()])
    if wname != "None":
        indc_dict = Dict(wmodel_type=indc_def_map[wname], induction=induction)
        algo_dict["mbook"].wake_models[wname] = WakeModel.new(**indc_dict)
        if verbosity > 2:
            print(f"      Created wake model '{wname}':")
            print("       ", algo_dict["mbook"].wake_models[wname])
        algo_dict["wake_models"].append(wname)
        algo_dict["algo_type"] = "Iterative"

def _read_rotor_averaging(rotor_averaging, algo_dict, verbosity):
    """Reads the rotor averaging"""
    if verbosity > 2:
        print("    Reading rotor_averaging")
        print("      Contents:", [k for k in rotor_averaging.keys()])
    grid = rotor_averaging["grid"]
    nx = rotor_averaging["n_x_grid_points"]
    ny = rotor_averaging["n_y_grid_points"]
    if nx != ny:
        raise NotImplementedError(
            f"Grid '{grid}': Only nx=ny supported, got nx={nx}, ny={ny}"
        )
    background_averaging = rotor_averaging["background_averaging"]
    wake_averaging = rotor_averaging["wake_averaging"]
    wse_P = rotor_averaging["wind_speed_exponent_for_power"]
    wse_ct = rotor_averaging["wind_speed_exponent_for_ct"]
    if verbosity > 2:
        print("        grid                :", grid)
        print("        background_averaging:", background_averaging)
        print("        wake_averaging      :", wake_averaging)
        print("        ws exponent power   :", wse_P)
        print("        ws exponent ct      :", wse_ct)
    if background_averaging in ["center", "centre"]:
        algo_dict["rotor_model"] = "centre"
    elif background_averaging == "grid":
        algo_dict["rotor_model"] = f"grid{nx*ny}"
    else:
        raise KeyError(
            f"Expecting background_averaging 'center' or 'grid', got '{background_averaging}'"
        )
    if wake_averaging in ["centre", "center"]:
        algo_dict["partial_wakes"] = "centre"
    elif wake_averaging == "grid":
        if background_averaging == "grid":
            algo_dict["partial_wakes"] = "rotor_points"
        else:
            if grid == "grid":
                algo_dict["partial_wakes"] = f"grid{nx*ny}"
            else:
                algo_dict["partial_wakes"] = grid
    else:
        algo_dict["partial_wakes"] = wake_averaging
    if verbosity > 2:
        print("        --> rotor_model     :", algo_dict["rotor_model"])
        print("        --> partial_wakes   :", algo_dict["partial_wakes"])

def _read_deflection(deflection, induction, algo_dict, verbosity):
    """Reads deflection model"""
    defl_def_map = Dict(
        {
            "None": "RotorWD",
            "Batankhah2016": "YawedWakes",
        },
        name="defl_def_map",
    )

    wname = deflection.pop("name")
    if verbosity > 2:
        print("    Reading deflection_model")
        print("      Name:", wname)
        print("      Contents:", [k for k in deflection.keys()])
    indc_dict = Dict(wframe_type=defl_def_map[wname])
    try:
        algo_dict["mbook"].wake_frames[wname] = WakeFrame.new(
            **indc_dict, induction=induction
        )
    except TypeError:
        algo_dict["mbook"].wake_frames[wname] = WakeFrame.new(**indc_dict)
    if verbosity > 2:
        print(f"      Created wake frame '{wname}':")
        print("       ", algo_dict["mbook"].wake_frames[wname])
    algo_dict["wake_frame"] = wname

def _read_analysis(wio_ana, algo_dict, verbosity):
    """Reads the windio analyses"""
    if verbosity > 2:
        print("    Reading analysis")
        print("      Contents:", [k for k in wio_ana.keys()])

    # superposition:
    superposition = Dict(wio_ana["superposition_model"], name="superposition_model")
    if verbosity > 2:
        print("    Reading superposition_model")
        print("      Contents:", [k for k in superposition.keys()])

    # axial induction model:
    imap = Dict(
        {
            "1D": "Betz",
            "Madsen": "Madsen",
        },
        name="induction mapping",
    )
    induction = imap[wio_ana.get("axial_induction_model", "1D")]
    if verbosity > 2:
        print("    axial induction model:", induction)

    # wind deficit model:
    wind_deficit = Dict(wio_ana["wind_deficit_model"], name="wind_deficit_model")
    ka, kb, amb_ti = _read_wind_deficit(
        wind_deficit, superposition, induction, algo_dict, verbosity
    )

    # turbulence model:
    if "turbulence_model" in wio_ana:
        turbulence_model = Dict(wio_ana["turbulence_model"], name="turbulence_model")
        _read_turbulence(
            turbulence_model, superposition, induction, algo_dict, ka, kb, amb_ti, verbosity
        )
    elif verbosity > 0:
        print("turbulence_model not found, not using a TI wake model")

    # blockage model:
    if "blockage_model" in wio_ana:
        blockage_model = Dict(wio_ana["blockage_model"], name="blockage_model")
        _read_blockage(blockage_model, induction, algo_dict, verbosity)
    elif verbosity > 0:
        print("blockage_model not found, not using a turbine induction model")
        
    # rotor_averaging:
    if "rotor_averaging" in wio_ana:
        rotor_averaging = Dict(wio_ana["rotor_averaging"], name="rotor_averaging")
        _read_rotor_averaging(rotor_averaging, algo_dict, verbosity)
    elif verbosity > 0:
        print("rotor_averaging not found, using default settings")
        
    # deflection:
    if "deflection_model" in wio_ana:
        deflection = Dict(wio_ana["deflection_model"], name="deflection_model")
        _read_deflection(deflection, induction, algo_dict, verbosity)
    elif verbosity > 0:
        print("deflection_model not found, using default settings")
        
def read_attributes(wio, algo_dict, verbosity):
    """
    Reads the attributes part of windio

    Parameters
    ----------
    wio: dict
        The windio data
    algo_dict: dict
        The algorithm dictionary
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    out_dicts: list of dict
        The output dictionaries

    :group: input.windio

    """
    wio_attrs = Dict(wio["attributes"], name="attributes")
    if verbosity > 1:
        print("Reading attributes")
        print("  Contents:", [k for k in wio_attrs.keys()])

    # read flow model:
    if "flow_model" in wio_attrs:
        flow_model = Dict(wio_attrs["flow_model"], name="flow_model")
        fmname = flow_model.pop("name")
        if verbosity > 2:
            print("    Reading flow_model")
            print("      Name:", fmname)
            print("      Contents:", [k for k in flow_model.keys()])
        if fmname != "foxes":
            print(f"Running flow model 'foxes', overruling original choice '{fmname}'")

    # read analysis:
    wio_ana = Dict(wio_attrs["analysis"], name="analyses")
    _read_analysis(wio_ana, algo_dict, verbosity)

    # outputs:
    out_dicts = []
    if "outputs" in wio_attrs:
        outputs = Dict(wio_attrs["outputs"], name="outputs")
        out_dicts = read_outputs(outputs, algo_dict, verbosity)

    return out_dicts



from foxes.core import Algorithm, FarmDataModelList
from foxes.core import PointDataModel, PointDataModelList, FarmController
import foxes.models as fm
import foxes.variables as FV
import foxes.constants as FC
from . import models as mdls


class Downwind(Algorithm):
    """
    The downwind algorithm.

    The turbines are evaluated once, in the order
    that is calculated by the provided `TurbineOrder`
    object.

    Attributes
    ----------
    states: foxes.core.States
        The ambient states
    wake_models: dict
        The wake models. Key: wake model name,
        value: foxes.core.WakeModel
    rotor_model: foxes.core.RotorModel
        The rotor model, for all turbines
    wake_frame: foxes.core.WakeFrame
        The wake frame
    partial_wakes: dict
        The partial wakes mapping. Key: wake model name,
        value: foxes.core.PartialWakesModel
    ground_models: dict
        The ground models mapping. Key: wake model name,
        value: foxes.core.GroundModel
    farm_controller: foxes.core.FarmController
        The farm controller
    n_states: int
        The number of states

    :group: algorithms.downwind

    """

    DEFAULT_FARM_OUTPUTS = [
        FV.X,
        FV.Y,
        FV.H,
        FV.D,
        FV.AMB_WD,
        FV.AMB_REWS,
        FV.AMB_TI,
        FV.AMB_RHO,
        FV.AMB_P,
        FV.WD,
        FV.REWS,
        FV.YAW,
        FV.TI,
        FV.CT,
        FV.P,
        FV.ORDER,
        FV.WEIGHT,
    ]

    def __init__(
        self,
        farm,
        states,
        wake_models,
        rotor_model="centre",
        wake_frame="rotor_wd",
        partial_wakes=None,
        ground_models=None,
        farm_controller="basic_ctrl",
        chunks={FC.STATE: 1000, FC.POINT: 4000},
        mbook=None,
        dbook=None,
        verbosity=1,
    ):
        """
        Constructor.

        Parameters
        ----------
        farm: foxes.WindFarm
            The wind farm
        states: foxes.core.States
            The ambient states
        wake_models: list of str
            The wake models, applied to all turbines.
            Will be looked up in the model book
        rotor_model: str
            The rotor model, for all turbines. Will be
            looked up in the model book
        wake_frame: str
            The wake frame. Will be looked up in the
            model book
        partial_wakes: dict, list or str, optional
            The partial wakes mapping. Key: wake model name,
            value: partial wake model name
        ground_models: dict, list or str, optional
            The ground models mapping. Key: wake model name,
            value: ground model name
        farm_controller: str
            The farm controller. Will be
            looked up in the model book
        chunks: dict
            The chunks choice for running in parallel with dask,
            e.g. `{"state": 1000}` for chunks of 1000 states
        mbook: foxes.ModelBook, optional
            The model book
        dbook: foxes.DataBook, optional
            The data book, or None for default
        verbosity: int
            The verbosity level, 0 means silent

        """
        if mbook is None:
            mbook = fm.ModelBook()

        super().__init__(mbook, farm, chunks, verbosity, dbook)

        self.states = states
        self.n_states = None
        self.states_data = None

        self.rotor_model = self.mbook.rotor_models[rotor_model]
        self.rotor_model.name = rotor_model

        self.wake_frame = self.mbook.wake_frames[wake_frame]
        self.wake_frame.name = wake_frame

        self.wake_models = {}
        for w in wake_models:
            m = self.mbook.wake_models[w]
            m.name = w
            self.wake_models[w] = m

        def _set_wspecific(descr, target, values, deffunc, mbooks, checkw):
            if values is None:
                values = {}
            if isinstance(values, list) and len(values) == 1:
                values = values[0]
            if isinstance(values, str):
                for w in wake_models:
                    try:
                        pw = values
                        if checkw:
                            mbooks[pw].check_wmodel(self.wake_models[w], error=True)
                    except TypeError:
                        pw = deffunc(self.wake_models[w])
                    target[w] = mbooks[pw]
                    target[w].name = pw
            elif isinstance(values, list):
                for i, w in enumerate(wake_models):
                    if i >= len(values):
                        raise IndexError(
                            f"Not enough {descr} in list {values}, expecting {len(wake_models)}"
                        )
                    pw = values[i]
                    target[w] = mbooks[pw]
                    target[w].name = pw
            else:
                for w in wake_models:
                    if w in values:
                        pw = values[w]
                    else:
                        pw = deffunc(self.wake_models[w])
                    target[w] = mbooks[pw]
                    target[w].name = pw

        self.partial_wakes = {}
        _set_wspecific(
            descr="partial wakes",
            target=self.partial_wakes,
            values=partial_wakes,
            deffunc=mbook.default_partial_wakes,
            mbooks=self.mbook.partial_wakes,
            checkw=True,
        )

        self.ground_models = {}
        _set_wspecific(
            descr="ground models",
            target=self.ground_models,
            values=ground_models,
            deffunc=lambda w: "no_ground",
            mbooks=self.mbook.ground_models,
            checkw=False,
        )

        self.farm_controller = self.mbook.farm_controllers[farm_controller]
        self.farm_controller.name = farm_controller

    @classmethod
    def get_model(cls, name):
        """
        Get the algorithm specific model

        Parameters
        ----------
        name: str
            The model name

        Returns
        -------
        model: foxes.core.model
            The model

        """
        return getattr(mdls, name)

    def _print_deco(self, func_name, n_points=None):
        """
        Helper function for printing model names
        """
        if self.verbosity > 0:
            deco = "-" * 50
            print(f"\n{deco}")
            print(f"  Running {self.name}: {func_name}")
            print(deco)
            print(f"  n_states : {self.n_states}")
            print(f"  n_turbines: {self.n_turbines}")
            if n_points is not None:
                print(f"  n_points : {n_points}")
            print(deco)
            print(f"  states   : {self.states}")
            print(f"  rotor    : {self.rotor_model}")
            print(f"  controller: {self.farm_controller}")
            print(f"  wake frame: {self.wake_frame}")
            print(deco)
            print(f"  wakes:")
            for i, w in enumerate(self.wake_models.values()):
                print(f"    {i}) {w.name}: {w}")
            print(deco)
            print(f"  partial wakes:")
            for i, (w, p) in enumerate(self.partial_wakes.items()):
                print(f"    {i}) {w}: {p.name}, {p}")
            print(deco)
            print(f"  turbine models:")
            for i, m in enumerate(self.farm_controller.pre_rotor_models.models):
                print(f"    {i}) {m.name}: {m} [pre-rotor]")
            for i, m in enumerate(self.farm_controller.post_rotor_models.models):
                print(
                    f"    {i+len(self.farm_controller.pre_rotor_models.models)}) {m.name}: {m}"
                )
            print(deco)
            print()

    def _print_model_oder(self, mlist, calc_pars):
        """
        Helper function for printing model names
        """
        if self.verbosity > 0:
            deco = "-" * 50
            print(f"\n{deco}")
            print(f"  Model oder")
            print(f"{deco}")

            for i, m in enumerate(mlist.models):
                print(f"{i:02d}) {m.name}")
                if isinstance(m, FarmController):
                    if calc_pars[i]["pre_rotor"]:
                        for j, mm in enumerate(m.pre_rotor_models.models):
                            print(f"  {i:02d}.{j}) Pre-rotor: {mm.name}")
                    else:
                        for j, mm in enumerate(m.post_rotor_models.models):
                            print(f"  {i:02d}.{j}) Post-rotor: {mm.name}")

            print(deco)
            print()

    def init_states(self):
        """
        Initialize states, if needed.
        """
        if not self.states.initialized:
            self.states.initialize(self, self.verbosity)
            self.n_states = self.states.size()

    def all_models(self, with_states=True):
        """
        Return all models

        Parameters
        ----------
        with_states: bool
            Flag for including states

        Returns
        -------
        mdls: list of foxes.core.Model
            The list of models

        """
        mdls = [self.states] if with_states else []
        mdls += [
            self.rotor_model,
            self.farm_controller,
            self.wake_frame,
        ]
        mdls += list(self.wake_models.values())
        mdls += list(self.partial_wakes.values())

        return mdls

    def initialize(self):
        """
        Initializes the algorithm.
        """
        self.print(f"\nInitializing algorithm '{self.name}'")
        super().initialize()

        self.init_states()

        for m in self.all_models(with_states=False):
            m.initialize(self, self.verbosity)

    def _collect_farm_models(
        self,
        outputs,
        calc_parameters,
        ambient,
    ):
        """
        Helper function that creates model list
        """
        # prepare:
        calc_pars = []
        mlist = FarmDataModelList(models=[])
        mlist.name = f"{self.name}_calc"

        # 0) run pre-rotor turbine models via farm controller:
        mlist.models.append(self.farm_controller)
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
        calc_pars[-1]["pre_rotor"] = True

        # 1) set initial data:
        mlist.models.append(self.get_model("InitFarmData")())
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        # 2) calculate ambient rotor results:
        mlist.models.append(self.rotor_model)
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
        calc_pars[-1].update(
            {"store_rpoints": True, "store_rweights": True, "store_amb_res": True}
        )

        # 3) run post-rotor turbine models via farm controller:
        mlist.models.append(self.farm_controller)
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))
        calc_pars[-1]["pre_rotor"] = False

        # 4) copy results to ambient, requires self.farm_vars:
        self.farm_vars = mlist.output_farm_vars(self)
        mlist.models.append(self.get_model("SetAmbFarmResults")())
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        # 5) calculate wake effects:
        if not ambient:
            mlist.models.append(self.get_model("FarmWakesCalculation")())
            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        # 6) reorder back to state-turbine dimensions:
        if outputs != False:
            mlist.models.append(self.get_model("ReorderFarmOutput")(outputs))
            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        return mlist, calc_pars

    def _calc_farm_vars(self, mlist):
        """Helper function that gathers the farm variables"""
        self.farm_vars = sorted(list(set([FV.WEIGHT] + mlist.output_farm_vars(self))))

    def _run_farm_calc(self, mlist, *data, outputs=None, **kwargs):
        """Helper function for running the main farm calculation"""
        self.print(
            f"\nCalculating {self.n_states} states for {self.n_turbines} turbines"
        )
        out_vars = self.farm_vars if outputs is None else outputs
        farm_results = mlist.run_calculation(self, *data, out_vars=out_vars, **kwargs)
        farm_results[FC.TNAME] = ((FC.TURBINE,), self.farm.turbine_names)
        for v in [FV.ORDER, FV.ORDER_SSEL, FV.ORDER_INV]:
            if v in farm_results:
                farm_results[v] = farm_results[v].astype(FC.ITYPE)

        return farm_results

    def calc_farm(
        self,
        outputs=None,
        calc_parameters={},
        persist=True,
        finalize=True,
        ambient=False,
        chunked_results=False,
        **kwargs,
    ):
        """
        Calculate farm data.

        Parameters
        ----------
        calc_parameters: dict
            Parameters for model calculation.
            Key: model name str, value: parameter dict
        outputs: list of str, optional
            The output variables, or None for defaults
        persist: bool
            Switch for forcing dask to load all model data
            into memory
        finalize: bool
            Flag for finalization after calculation
        ambient: bool
            Flag for ambient instead of waked calculation
        chunked_results: bool
            Flag for chunked results
        kwargs: dict, optional
            Additional parameters for run_calculation

        Returns
        -------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)

        """
        # initialize algorithm:
        if not self.initialized:
            self.initialize()

        # welcome:
        self._print_deco("calc_farm")

        # collect models:
        if outputs == "default":
            outputs = self.DEFAULT_FARM_OUTPUTS
        mlist, calc_pars = self._collect_farm_models(outputs, calc_parameters, ambient)

        # initialize models:
        if not mlist.initialized:
            mlist.initialize(self, self.verbosity)
            self._calc_farm_vars(mlist)
        self._print_model_oder(mlist, calc_pars)

        # update outputs:
        if outputs is None:
            outputs = self.farm_vars
        else:
            outputs = sorted(list(set(outputs).intersection(self.farm_vars)))

        # get input model data:
        models_data = self.get_models_data()
        if persist:
            models_data = models_data.persist()
        self.print("\nInput data:\n\n", models_data, "\n")
        self.print(f"\nFarm variables:", ", ".join(self.farm_vars))
        self.print(f"\nOutput variables:", ", ".join(outputs))
        self.print(f"\nChunks: {self.chunks}\n")

        # run main calculation:
        farm_results = self._run_farm_calc(
            mlist,
            models_data,
            parameters=calc_pars,
            outputs=outputs,
            **kwargs,
        )
        del models_data

        # finalize models:
        if finalize:
            self.print("\n")
            mlist.finalize(self, self.verbosity)
            self.finalize()

        if ambient:
            dvars = [v for v in farm_results.data_vars.keys() if v in FV.var2amb]
            farm_results = farm_results.drop_vars(dvars)

        if chunked_results:
            farm_results = self.chunked(farm_results)

        return farm_results

    def _collect_point_models(
        self,
        calc_parameters={},
        point_models=None,
        ambient=False,
    ):
        """
        Helper function that creates model list
        """
        # prepare:
        calc_pars = []
        mlist = PointDataModelList(models=[])

        # prepare extra eval models:
        emodels = []
        emodels_cpars = []
        if point_models is not None:
            if not isinstance(point_models, list):
                point_models = [point_models]
            for m in point_models:
                if isinstance(m, str):
                    pname = m
                    pmodel = self.mbook.point_models[pname]
                    pmodel.name = pname
                    emodels.append(pmodel)
                elif isinstance(m, PointDataModel):
                    emodels.append(m)
                else:
                    raise TypeError(f"Model '{m}' is neither str nor PointDataModel")
                emodels_cpars.append(calc_parameters.get(emodels[-1].name, {}))
        emodels = PointDataModelList(models=emodels)

        # 0) calculate states results:
        mlist.models.append(self.states)
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        # 1) calculate ambient extra eval point results:
        mlist.models.append(emodels)
        calc_pars.append({"parameters": emodels_cpars})

        # 2) transfer ambient results:
        mlist.models.append(self.get_model("SetAmbPointResults")())
        calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        # 3) calc wake effects:
        if not ambient:
            mlist.models.append(
                self.get_model("PointWakesCalculation")(emodels, emodels_cpars)
            )
            calc_pars.append(calc_parameters.get(mlist.models[-1].name, {}))

        return mlist, calc_pars

    def calc_points(
        self,
        farm_results,
        points,
        point_models=None,
        calc_parameters={},
        persist_mdata=True,
        persist_pdata=False,
        finalize=True,
        ambient=False,
        chunked_results=False,
        **kwargs,
    ):
        """
        Calculate data at a given set of points.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)
        points: numpy.ndarray
            The points of interest, shape: (n_states, n_points, 3)
        point_models: str or foxes.core.PointDataModel
            Additional point models to be executed
        calc_parameters: dict
            Parameters for model calculation.
            Key: model name str, value: parameter dict
        persist_mdata: bool
            Switch for forcing dask to load all model data
            into memory
        persist_fdata: bool
            Switch for forcing dask to load all farm data
            into memory
        finalize: bool
            Flag for finalization after calculation
        ambient: bool
            Flag for ambient instead of waked calculation
        chunked_results: bool
            Flag for chunked results
        kwargs: dict, optional
            Additional parameters for run_calculation

        Returns
        -------
        point_results: xarray.Dataset
            The point results. The calculated variables have
            dimensions (state, point)

        """

        if not self.initialized:
            self.initialize()
        if not ambient and farm_results is None:
            raise ValueError(
                f"Cannot calculate point results without farm results for ambient = {ambient}"
            )

        # welcome:
        self._print_deco("calc_points", n_points=points.shape[1])

        # collect models and initialize:
        mlist, calc_pars = self._collect_point_models(
            calc_parameters, point_models, ambient
        )

        # initialize models:
        if not mlist.initialized:
            mlist.initialize(self, self.verbosity)

        # get input model data:
        models_data = self.get_models_data()
        if persist_mdata:
            models_data = models_data.persist()
        self.print("\nInput data:\n\n", models_data, "\n")
        self.print(f"\nOutput farm variables:", ", ".join(self.farm_vars))
        self.print(f"\nChunks: {self.chunks}\n")

        # chunk farm results:
        if self.chunks is not None:
            farm_results = self.chunked(farm_results)
        self.print("\nInput farm data:\n\n", farm_results, "\n")

        # get point data:
        if FC.STATE in farm_results.coords:
            sinds = farm_results.coords[FC.STATE]
        elif models_data is not None and FC.STATE in models_data.coords:
            sinds = models_data.coords[FC.STATE]
        else:
            sinds = None
        point_data = self.new_point_data(points, sinds)
        if persist_pdata:
            point_data = point_data.persist()
        self.print("\nInput point data:\n\n", point_data, "\n")

        # check vars:
        ovars = mlist.output_point_vars(self)
        self.print(f"\nOutput point variables:", ", ".join(ovars))
        self.print(f"\nChunks: {self.chunks}\n")

        # calculate:
        self.print(
            f"Calculating {len(ovars)} variables at {points.shape[1]} points in {self.n_states} states"
        )

        point_results = mlist.run_calculation(
            self,
            models_data,
            farm_results,
            point_data,
            out_vars=ovars,
            parameters=calc_pars,
            **kwargs,
        )

        del models_data, farm_results, point_data

        # finalize models:
        if finalize:
            self.print("\n")
            mlist.finalize(self, self.verbosity)
            self.finalize()

        if ambient:
            dvars = [v for v in point_results.data_vars.keys() if v in FV.var2amb]
            point_results = point_results.drop_vars(dvars)

        if chunked_results:
            point_results = self.chunked(point_results)

        return point_results

    def finalize(self, clear_mem=False):
        """
        Finalizes the algorithm.

        Parameters
        ----------
        clear_mem: bool
            Clear idata memory

        """
        for m in self.all_models():
            m.finalize(self, self.verbosity)

        super().finalize(clear_mem)



import numpy as np
import pandas as pd

import foxes.variables as FV


def random_timseries_data(
    n_times,
    data_ranges=None,
    start_time="2000-01-01 00:00:00",
    freq="h",
    seed=None,
    iname="Time",
):
    """
    Creates random uniform timeseries data

    Parameters
    ----------
    n_times: int
        The number of time steps
    data_ranges: dict, optional
        The data ranges. Key: variable name,
        value: tuple, [min, max) values
    start_time: str
        The first time stamp in the series
    freq: str
        The time period range frequency
    seed: int, optional
        The random seed
    iname: str
        The index name

    Returns
    -------
    sdata: pandas.DataFrame
        The timeseries data

    :group: input.states.create

    """
    if seed:
        np.random.seed(seed)

    dranges = {FV.WS: (0.0, 30.0), FV.WD: (0.0, 360.0)}
    if data_ranges:
        dranges.update(data_ranges)

    times = pd.period_range(start=start_time, periods=n_times, freq=freq)
    times = times.astype(str).astype("datetime64[ns]")
    sdata = pd.DataFrame(
        index=times,
        data={v: np.random.uniform(d[0], d[1], n_times) for v, d in dranges.items()},
    )
    sdata.index.name = iname
    return sdata



import numpy as np

from foxes.core import VerticalProfile


class UniformProfile(VerticalProfile):
    """
    A profile with uniform values.

    Attributes
    ----------
    var: float
        The value

    :group: models.vertical_profiles

    """

    def __init__(self, variable):
        """
        Constructor

        Parameters
        ----------
        variable: float
            The value

        """
        super().__init__()
        self.var = variable

    def input_vars(self):
        """
        The input variables needed for the profile
        calculation.

        Returns
        -------
        vars: list of str
            The variable names

        """
        return [self.var]

    def calculate(self, data, heights):
        """
        Run the profile calculation.

        Parameters
        ----------
        data: dict
            The input data
        heights: numpy.ndarray
            The evaluation heights

        Returns
        -------
        results: numpy.ndarray
            The profile results, same
            shape as heights

        """
        out = np.zeros_like(heights)
        out[:] = data[self.var]
        return out



import numpy as np
import argparse

import foxes.constants as FC
from .dict import Dict
from .geom2d import AreaUnion, ClosedPolygon

try:
    import geopandas as gpd

    IMPORT_GPD_OK = True
except ImportError:
    gpd = None
    IMPORT_GPD_OK = False

try:
    import utm

    IMPORT_UTM_OK = True
except ImportError:
    utm = None
    IMPORT_UTM_OK = False


def check_import_gpd():
    """
    Checks if library import worked,
    raises error otherwise.
    """
    if not IMPORT_GPD_OK:
        print("\n\nFailed to import geopandas. Please install, either via pip:\n")
        print("  pip install geopandas\n")
        print("or via conda:\n")
        print("  conda install -c conda-forge geopandas\n")
        raise ImportError("Failed to import geopandas")


def check_import_utm():
    """
    Checks if library import worked,
    raises error otherwise.
    """
    if not IMPORT_UTM_OK:
        print("\n\nFailed to import utm. Please install, either via pip:\n")
        print("  pip install utm\n")
        print("or via conda:\n")
        print("  conda install -c conda-forge utm\n")
        raise ImportError("Failed to import utm")


def read_shp(fname, **kwargs):
    """
    Read a shape file

    Parameters
    ----------
    fname: str
        Path to the .shp file
    kwargs: dict, optional
        Additional parameters for geopandas.read_file()

    Returns
    -------
    data: geopandas.GeoDataFrame
        The data frame in WSG84

    :group: utils

    """
    check_import_gpd()
    gpdf = gpd.read_file(fname, **kwargs)
    return gpdf.to_crs("EPSG:4326")  # Convert to WGS84


def shp2csv(ifile, ofile, in_kwargs={}, out_kwargs={}, verbosity=1):
    """
    Read shape file, write csv file

    Parameters
    ----------
    iname: str
        Path to the input .shp file
    oname: str
        Path to the output .csv file
    in_kwargs: dict
        Additional parameters for geopandas.read_file()
    out_kwargs: dict
        Additional parameters for geopandas to_csv()
    verbosity: int
        The verbosity level, 0 = silent

    :group: utils

    """
    if verbosity > 0:
        print("Reading file", ifile)

    gpdf = read_shp(ifile, **in_kwargs)

    if verbosity > 0:
        print("Writing file", ofile)

    gpdf.to_csv(ofile, **out_kwargs)

    return gpdf


def _extract_poly_coords(geom):
    """
    Helper function for shapefile reading
    """
    if geom.geom_type == "Polygon":
        exterior_coords = geom.exterior.coords[:]
        interior_coords = []
        for interior in geom.interiors:
            interior_coords.append(interior.coords[:])
    elif geom.geom_type == "MultiPolygon":
        exterior_coords = []
        interior_coords = []
        for part in geom.geoms:
            epe, epi = _extract_poly_coords(part)  # Recursive call
            exterior_coords.append(epe)
            interior_coords.append(epi)
    else:
        raise ValueError("Unhandled geometry type: " + repr(geom.type))
    return exterior_coords, interior_coords


def read_shp_polygons(
    fname,
    names=None,
    name_col="Name",
    geom_col="geometry",
    to_utm=True,
    ret_utm_zone=False,
    **kwargs,
):
    """
    Reads the polygon points from a shp file.

    Parameters
    ----------
    fname: str
        Path to the .shp file
    names: list: of str, optinal
        The names of the polygons to be extracted. All by
        default
    name_col: int
        Column that contains the area names
    geom_col: str
        The geometry column
    to_utm: bool or str, optional
        Convert to UTM coordinates. If str, then UTM zone
        plus letter, e.g. "32U"
    ret_utm_zone: bool
        Return UTM zone plus letter as str
    kwargs: dict, optional
        Additional parameters for geopandas.read_shp()

    Returns
    -------
    point_dict_exterior: dict
        Dict with list of array of points. Key: area name,
        Value: list:np.ndarray, shape of latter: (n_points, 2)
    point_dict_interior: dict
        Dict with list of array of points. Key: area name,
        Value: list:np.ndarray, shape of latter: (n_points, 2)
    utm_zone_str: str, optional
        The utem zone plus letter as str, e.g. "32U"

    :group: utils

    """

    pdf = read_shp(fname, **kwargs)
    pnames = list(pdf[name_col])

    utmz = None
    utml = None
    apply_utm = False
    if isinstance(to_utm, str) or to_utm == True:
        apply_utm = True
        check_import_utm()
        utmz = int(to_utm[:-1]) if isinstance(to_utm, str) else None
        utml = to_utm[-1] if isinstance(to_utm, str) else None

    exterior = Dict()
    interior = Dict()
    names = pnames if names is None else names
    for name in names:
        if name == name:  # exclude nan values
            if not name in pnames:
                raise KeyError(
                    f"Name '{name}' not found in file '{fname}'. Names: {pnames}"
                )

            a = pdf.loc[pnames.index(name), geom_col]
            epe, epi = _extract_poly_coords(a)

            def _to_utm(poly):
                nonlocal utmz, utml
                utm_poly = np.zeros_like(poly)
                utm_poly[:, 0], utm_poly[:, 1], utmz, utml = utm.from_latlon(
                    poly[:, 1],
                    poly[:, 0],
                    force_zone_number=utmz,
                    force_zone_letter=utml,
                )
                return utm_poly

            def _to_numpy(data):
                if not len(data):
                    return []
                if isinstance(data[0], tuple):
                    out = np.array(data, dtype=FC.DTYPE)
                    return _to_utm(out) if apply_utm else out
                return [_to_numpy(d) for d in data]

            exterior[name] = _to_numpy(epe)
            interior[name] = _to_numpy(epi)

    if ret_utm_zone:
        return exterior, interior, f"{utmz}{utml}"
    else:
        return exterior, interior


def shp2geom2d(*args, ret_utm_zone=False, **kwargs):
    """
    Read shapefile into geom2d geometry

    Parameters
    ----------
    args: tuple, optional
        Arguments for read_shp_polygons()
    ret_utm_zone: bool
        Return UTM zone plus letter as str
    kwargs: dict, optional
        Keyword arguments for read_shp_polygons()

    Returns
    -------
    geom: foxes.tools.geom2D.AreaGeometry
        The geometry object
    utm_zone_str: str, optional
        The utem zone plus letter as str, e.g. "32U"

    :group: utils

    """

    exint = read_shp_polygons(*args, ret_utm_zone=ret_utm_zone, **kwargs)

    def _create_geom(data):
        if not len(data):
            return None
        if isinstance(data, dict):
            gs = [_create_geom(g) for g in data.values()]
            gs = [g for g in gs if g is not None]
            return AreaUnion(gs) if len(gs) else None
        if isinstance(data, np.ndarray) and len(data.shape) == 2:
            return ClosedPolygon(data)
        gs = [_create_geom(g) for g in data]
        gs = [g for g in gs if g is not None]
        return AreaUnion(gs) if len(gs) else None

    gext = _create_geom(exint[0])
    gint = _create_geom(exint[1])
    geom = gext - gint if gint is not None else gext

    if ret_utm_zone:
        return geom, exint[2]
    else:
        return geom


if __name__ == "__main__":
    # define arguments and options:
    parser = argparse.ArgumentParser()
    parser.add_argument("shp_file", help="The input .shp file")
    parser.add_argument("-n", "--names", help="Area names", default=None, nargs="+")
    parser.add_argument(
        "--no_utm", help="switch off conversion to UTM", action="store_true"
    )
    args = parser.parse_args()

    g = shp2geom2d(args.shp_file, to_utm=not args.no_utm, names=args.names)

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    g.add_to_figure(ax)
    plt.show()
    plt.close(fig)



from foxes.utils import DataBook

from . import farms
from . import states
from . import power_ct_curves
from . import windio

FARM = "farm"
""" Static wind farm data identifier
:group: data
"""

STATES = "states"
""" Static states data identifier
:group: data
"""

PCTCURVE = "power_ct_curve"
""" Static power-ct curve data identifier
:group: data
"""

WINDIO = "windio"
""" Static windio data identifier
:group: data
"""


class StaticData(DataBook):
    """
    A DataBook filled with static data from
    this directory.

    :group: data

    """

    def __init__(self):
        super().__init__()

        self.add_data_package(FARM, farms, ".csv")
        self.add_data_package(STATES, states, [".csv", ".csv.gz", ".nc", ".tab"])
        self.add_data_package(PCTCURVE, power_ct_curves, ".csv")
        self.add_data_package(WINDIO, windio, ".yaml")



import numpy as np
from scipy.interpolate import interpn

from foxes.core import WakeFrame
from foxes.utils import wd2uv
from foxes.core.data import TData
import foxes.variables as FV
import foxes.constants as FC


class Streamlines2D(WakeFrame):
    """
    Streamline following wakes

    Attributes
    ----------
    step: float
        The streamline step size in m
    max_length: float
        The maximal streamline length
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation

    :group: models.wake_frames

    """

    def __init__(self, step, max_length=1e4, cl_ipars={}):
        """
        Constructor.

        Parameters
        ----------
        step: float
            The streamline step size in m
        max_length: float
            The maximal streamline length
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation

        """
        super().__init__()
        self.step = step
        self.max_length = max_length
        self.cl_ipars = cl_ipars

        self.DATA = self.var("DATA")
        self.STEPS = self.var("STEPS")
        self.SDAT = self.var("SDAT")

    def __repr__(self):
        return f"{type(self).__name__}(step={self.step})"

    def _calc_streamlines(self, algo, mdata, fdata):
        """
        Helper function that computes all streamline data
        """
        # prepare:
        n_states = mdata.n_states
        n_turbines = mdata.n_turbines
        N = int(self.max_length / self.step)

        # calc data: x, y, z, wd
        data = np.zeros((n_states, n_turbines, N, 4), dtype=FC.DTYPE)
        for i in range(N):

            # set streamline start point data (rotor centre):
            if i == 0:
                data[:, :, i, :3] = fdata[FV.TXYH]
                data[:, :, i, 3] = fdata[FV.AMB_WD]

            # compute next step:
            else:

                # calculate next point:
                xyz = data[:, :, i - 1, :3]
                n = wd2uv(data[:, :, i - 1, 3])
                data[:, :, i, :2] = xyz[:, :, :2] + self.step * n
                data[:, :, i, 2] = xyz[:, :, 2]

                # calculate next tangential vector:
                svars = algo.states.output_point_vars(algo)
                tdata = TData.from_points(
                    data[:, :, i, :3],
                    data={
                        v: np.full((n_states, n_turbines, 1), np.nan, dtype=FC.DTYPE)
                        for v in svars
                    },
                    dims={v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in svars},
                )
                data[:, :, i, 3] = algo.states.calculate(algo, mdata, fdata, tdata)[
                    FV.WD
                ][:, :, 0]

                sel = np.isnan(data[:, :, i, 3])
                if np.any(sel):
                    data[sel, i, 3] = data[sel, i - 1, 3]

        return data

    def get_streamline_data(self, algo, mdata, fdata):
        """
        Gets streamline data, generating it on the fly

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        data: numpy.ndarray
            The streamline data, shape:
            (n_states, n_turbines, n_steps, 4)
            with data x, y, z, wd

        """
        if self.DATA not in mdata or not np.all(
            mdata[self.DATA][:, :, 0, :3] == fdata[FV.TXYH]
        ):
            mdata[self.DATA] = self._calc_streamlines(algo, mdata, fdata)
            mdata.dims[self.DATA] = (FC.STATE, FC.TURBINE, self.STEPS, self.SDAT)

        return mdata[self.DATA]

    def _calc_coos(self, algo, mdata, fdata, targets, downwind_index):
        """
        Helper function, calculates streamline coordinates
        for given points and given turbine
        """

        # prepare:
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)

        # find nearest streamline points:
        data = self.get_streamline_data(algo, mdata, fdata)[:, downwind_index]
        dists = np.linalg.norm(points[:, :, None, :2] - data[:, None, :, :2], axis=-1)
        selp = np.argmin(dists, axis=2)
        data = np.take_along_axis(data[:, None], selp[:, :, None, None], axis=2)[
            :, :, 0
        ]
        slen = self.step * selp
        del dists, selp

        # calculate coordinates:
        coos = np.zeros((n_states, n_points, 3), dtype=FC.DTYPE)
        nx = wd2uv(data[:, :, 3])
        ny = np.stack([-nx[:, :, 1], nx[:, :, 0]], axis=2)
        delta = points[:, :, :2] - data[:, :, :2]
        coos[:, :, 0] = slen + np.einsum("spd,spd->sp", delta, nx)
        coos[:, :, 1] = np.einsum("spd,spd->sp", delta, ny)
        coos[:, :, 2] = points[:, :, 2] - data[:, :, 2]

        return coos.reshape(n_states, n_targets, n_tpoints, 3)

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines
        tdata = TData.from_points(points=fdata[FV.TXYH])

        # calculate streamline x coordinates for turbines rotor centre points:
        # n_states, n_turbines_source, n_turbines_target
        coosx = np.zeros((n_states, n_turbines, n_turbines), dtype=FC.DTYPE)
        for ti in range(n_turbines):
            coosx[:, ti, :] = self.get_wake_coos(algo, mdata, fdata, tdata, ti)[
                :, :, 0, 0
            ]

        # derive turbine order:
        # TODO: Remove loop over states
        order = np.zeros((n_states, n_turbines), dtype=FC.ITYPE)
        for si in range(n_states):
            order[si] = np.lexsort(keys=coosx[si])

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        return self._calc_coos(algo, mdata, fdata, tdata[FC.TARGETS], downwind_index)

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        # calculate long enough streamlines:
        xmax = np.max(x)
        self._ensure_min_length(algo, mdata, fdata, xmax)

        # get streamline points:
        n_states, n_points = x.shape
        data = self.get_streamline_data(algo, mdata, fdata)[:, downwind_index]
        spts = data[:, :, :3]
        n_spts = spts.shape[1]
        xs = self.step * np.arange(n_spts)

        # interpolate to x of interest:
        qts = np.zeros((n_states, n_points, 2), dtype=FC.DTYPE)
        qts[:, :, 0] = np.arange(n_states)[:, None]
        qts[:, :, 1] = np.minimum(x, xs[-1])
        qts = qts.reshape(n_states * n_points, 2)
        ipars = dict(bounds_error=False, fill_value=0.0)
        ipars.update(self.cl_ipars)
        results = interpn((np.arange(n_states), xs), spts, qts, **ipars)

        return results.reshape(n_states, n_points, 3)



import numpy as np

from foxes.core import States, Data
import foxes.constants as FC
import foxes.variables as FV


class PopStates(States):
    """
    Helper class for vectorized opt population
    calculation, via artificial states of length
    n_pop times n_states.

    Attributes
    ----------
    states: foxes.core.States
        The original states
    n_pop: int
        The population size

    :group: opt.core

    """

    def __init__(self, states, n_pop):
        """
        Constructor.

        Parameters
        ----------
        states: foxes.core.States
            The original states
        n_pop: int
            The population size

        """
        super().__init__()
        self.states = states
        self.n_pop = n_pop

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self.STATE0 = self.var(FC.STATE + "0")
        self.SMAP = self.var("SMAP")

        idata = super().load_data(algo, verbosity)
        idata0 = algo.get_model_data(self.states)
        for cname, coord in idata0["coords"].items():
            if cname != FC.STATE:
                idata["coords"][cname] = coord
            else:
                idata["coords"][self.STATE0] = coord

        for dname, (dims0, data0) in idata0["data_vars"].items():
            if dname != FV.WEIGHT:
                hdims = tuple([d if d != FC.STATE else self.STATE0 for d in dims0])
                idata["data_vars"][dname] = (hdims, data0)

        smap = np.zeros((self.n_pop, self.states.size()), dtype=np.int32)
        smap[:] = np.arange(self.states.size())[None, :]
        smap = smap.reshape(self.size())
        idata["data_vars"][self.SMAP] = ((FC.STATE,), smap)

        found = False
        for dname, (dims0, data0) in idata["data_vars"].items():
            if self.STATE0 in dims0:
                found = True
                break
        if not found:
            del idata["coords"][self.STATE0]

        return idata

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not self.states.initialized:
            self.states.initialize(algo, verbosity)
        super().initialize(algo, verbosity)

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self.states.size() * self.n_pop

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        weights = np.zeros(
            (self.n_pop, self.states.size(), algo.n_turbines), dtype=FC.DTYPE
        )
        weights[:] = self.states.weights(algo)[None, :, :] / self.n_pop
        return weights.reshape(self.size(), algo.n_turbines)

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.states.output_point_vars(algo)

    def calculate(self, algo, mdata, fdata, pdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        pdata: foxes.core.Data
            The point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """

        hdata = {}
        hdims = {}
        smap = mdata[self.SMAP]
        for dname, data in mdata.items():
            dms = mdata.dims[dname]
            if dname == self.SMAP or dname == self.STATE0:
                pass
            elif dms[0] == self.STATE0:
                hdata[dname] = data[smap]
                hdims[dname] = tuple([FC.STATE] + list(dms)[1:])
            elif self.STATE0 in dms:
                raise ValueError(
                    f"States '{self.name}': Found states variable not at dimension 0 for mdata entry '{dname}': {dms}"
                )
            else:
                hdata[dname] = data
                hdims[dname] = dms
        hmdata = Data(hdata, hdims, mdata.loop_dims)

        return self.states.calculate(algo, hmdata, fdata, pdata)



import numpy as np
import matplotlib.pyplot as plt

from .circle import Circle
from .polygon import ClosedPolygon
from .area_geometry import AreaIntersection

if __name__ == "__main__":
    centres = []
    radii = []
    N = 500

    plist = [
        np.array([[1.0, 1.0], [1.3, 6], [5.8, 6.2], [6.5, 0.8]]),
        np.array([[1.5, 1.5], [1.5, 8], [2.5, 8], [2.5, 1.5]]),
    ]

    circles = [Circle(centres[i], radii[i]) for i in range(len(centres))]
    polygons = [ClosedPolygon(plist[i]) for i in range(len(plist))]

    fig, ax = plt.subplots()
    g = AreaIntersection(circles + polygons)
    g.add_to_figure(ax, show_boundary=True)
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaIntersection(circles + polygons)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaIntersection(circles + polygons)
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaIntersection(circles + polygons).inverse()
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = AreaIntersection(circles + polygons).inverse()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = polygons[0] - polygons[1]
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = polygons[0] - polygons[1]
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)



import numpy as np
import pandas as pd

from .PCt_from_two import PCtFromTwo
from foxes.data import parse_Pct_two_files
from foxes.utils import PandasFileHelper
import foxes.constants as FC


class CpCtFromTwo(PCtFromTwo):
    """
    Calculate power and ct by interpolating
    cp and ct from two files (or two pandas
    DataFrames).

    :group: models.turbine_types

    """

    def __init__(
        self,
        data_source_cp,
        data_source_ct,
        col_ws_cp_file="ws",
        col_cp="cp",
        rho=1.225,
        pd_file_read_pars_cp={},
        pd_file_read_pars_ct={},
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source_cp: str or pandas.DataFrame
            The file path, static name, or data
        data_source_ct: str or pandas.DataFrame
            The file path, static name, or data
        col_ws_cp_file: str
            The wind speed column in the file of the cp curve
        col_cp: str
            The cp column
        rho: float
            The air density for the curves
        pd_file_read_pars_cp:  dict
            Parameters for pandas cp file reading
        pd_file_read_pars_ct:  dict
            Parameters for pandas ct file reading
        parameters: dict, optional
            Additional parameters for PCtFromTwo class

        """
        if not isinstance(data_source_cp, pd.DataFrame) or not isinstance(
            data_source_ct, pd.DataFrame
        ):
            pars = parse_Pct_two_files(data_source_cp, data_source_ct)
            data_cp = PandasFileHelper.read_file(data_source_cp, **pd_file_read_pars_cp)
            data_ct = PandasFileHelper.read_file(data_source_ct, **pd_file_read_pars_ct)
        else:
            data_cp = data_source_cp
            data_ct = data_source_ct
            pars = parameters

        D = pars["D"]
        A = np.pi * (D / 2) ** 2
        ws = data_cp[col_ws_cp_file].to_numpy()
        cp = data_cp[col_cp].to_numpy()
        P_unit = pars.pop("P_unit", FC.kW)

        ws_delta = 0.0001
        ws_min = np.min(ws)
        ws_max = np.max(ws)
        N = int((ws_max - ws_min) / ws_delta)

        data_P = pd.DataFrame(index=range(N), dtype=FC.DTYPE)
        data_P["ws"] = np.linspace(ws_min, ws_max, N, endpoint=True)
        data_P["cp"] = np.interp(data_P["ws"], ws, cp, left=0, right=0)
        data_P["P"] = (
            0.5 * rho * A * data_P["cp"] * data_P["ws"] ** 3 / FC.P_UNITS[P_unit]
        )

        super().__init__(
            data_P,
            data_ct,
            col_ws_P_file="ws",
            col_P="P",
            rho=rho,
            P_unit=P_unit,
            **pars,
        )



class SequentialPlugin:
    """
    Base class for plugins that are
    updated with each sequential iteration

    Parameters
    ----------
    algo: foxes.algorithms.sequential.Sequential
        The sequetial algorithm

    :group: algorithms.sequential.models

    """

    def __init__(self):
        """
        Constructor.
        """
        self.algo = None

    def initialize(self, algo):
        """
        Initialize data based on the intial iterator

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The current sequetial algorithm

        """
        self.algo = algo

    def update(self, algo, fres, pres=None):
        """
        Updates data based on current iteration

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The latest sequetial algorithm
        fres: xarray.Dataset
            The latest farm results
        pres: xarray.Dataset, optional
            The latest point results

        """
        self.algo = algo

    def finalize(self, algo):
        """
        Finalize data based on the final iterator

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The final sequetial algorithm

        """
        self.algo = None



from foxes.core import PartialWakesModel


class RotorPoints(PartialWakesModel):
    """
    Partial wakes calculation directly by the
    rotor model.

    :group: models.partial_wakes

    """

    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        rotor = algo.rotor_model
        return (
            rotor.from_data_or_store(rotor.RPOINTS, algo, mdata),
            rotor.from_data_or_store(rotor.RWEIGHTS, algo, mdata),
        )

    def finalize_wakes(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        amb_res,
        rpoint_weights,
        wake_deltas,
        wmodel,
        downwind_index,
    ):
        """
        Updates the wake_deltas at the selected target
        downwind index.

        Modifies wake_deltas on the fly.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.Data
            The target point data
        amb_res: dict
            The ambient results at the target points
            of all rotors. Key: variable name, value
            np.ndarray of shape:
            (n_states, n_turbines, n_rotor_points)
        rpoint_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)
        wake_deltas: dict
            The wake deltas. Key: variable name,
            value: np.ndarray of shape
            (n_states, n_turbines, n_tpoints)
        wmodel: foxes.core.WakeModel
            The wake model
        downwind_index: int
            The index in the downwind order

        Returns
        -------
        final_wake_deltas: dict
            The final wake deltas at the selected downwind
            turbines. Key: variable name, value: np.ndarray
            of shape (n_states, n_rotor_points)

        """
        ares = {v: d[:, downwind_index, None] for v, d in amb_res.items()}
        wdel = {v: d[:, downwind_index, None].copy() for v, d in wake_deltas.items()}
        wmodel.finalize_wake_deltas(algo, mdata, fdata, ares, wdel)

        return {v: d[:, 0] for v, d in wdel.items()}



from xarray import Dataset

from foxes.utils import write_nc
import foxes.constants as FC 

from .output import Output

class StateTurbineTable(Output):
    """
    Creates tables of state-turbine type data
    
    Attributes
    ----------
    farm_results: xarray.Dataset
        The farm results
    
    :group: output
    
    """
    def __init__(self, farm_results):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset
            The farm results

        """
        self.farm_results = farm_results
    
    def get_dataset(
        self, 
        variables, 
        name_map={}, 
        to_file=None, 
        **kwargs,
        ):
        """
        Creates a dataset object
        
        Parameters
        ----------
        variables: list of str
            The output variables
        name_map: dict
            Map from foxes to output names
        to_file: str, optional
            The output file path, if writing is desired
        kwargs: dict, optional
            Additional parameters for write_nc
        
        Returns
        -------
        table: xarray.Dataset
            The state-turbine data table
            
        """
        state = name_map.get(FC.STATE, FC.STATE)
        turbine = name_map.get(FC.TURBINE, FC.TURBINE)

        ds = Dataset(
            coords={
                state: self.farm_results[FC.STATE].to_numpy(),
                turbine: self.farm_results[FC.TURBINE].to_numpy(),
            },
            data_vars={
                name_map.get(v, v): (
                    (state, turbine), self.farm_results[v].to_numpy())
                for v in variables
            }
        )
        
        if to_file is not None:
            write_nc(ds=ds, fpath=to_file, **kwargs)
        
        return ds
    


import numpy as np
from abc import ABCMeta, abstractmethod
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap


class AreaGeometry(metaclass=ABCMeta):
    """
    Abstract base class for closed 2D geometries.

    :group: utils.geom2d

    """

    @abstractmethod
    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        pass

    @abstractmethod
    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        pass

    @abstractmethod
    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """
        pass

    @abstractmethod
    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        pass

    def add_to_figure(
        self,
        ax,
        show_boundary=False,
        fill_mode="inside_slategray",
        pars_boundary={},
        pars_distance={},
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        if fill_mode is not None:
            if "Nx" in pars_distance or "Ny" in pars_distance:
                Nx = pars_distance.pop("Nx")
                Ny = pars_distance.pop("Ny")
            elif "N" in pars_distance:
                Nx = pars_distance.pop("N")
                Ny = Nx
            else:
                Nx = 500
                Ny = 500

            p0 = pars_distance.pop("p_min", self.p_min())
            p1 = pars_distance.pop("p_max", self.p_max())
            if np.isinf(p0[0]):
                q0 = self.inverse().p_min()
                a0 = ax.get_xlim()[0]
                p0[0] = a0 if a0 < q0[0] else q0[0]
            if np.isinf(p0[1]):
                q0 = self.inverse().p_min()
                a0 = ax.get_ylim()[0]
                p0[1] = a0 if a0 < q0[1] else q0[1]
            if np.isinf(p1[0]):
                q1 = self.inverse().p_max()
                a1 = ax.get_xlim()[1]
                p1[0] = a1 if a1 > q1[0] else q1[0]
            if np.isinf(p1[1]):
                q1 = self.inverse().p_max()
                a1 = ax.get_ylim()[1]
                p1[1] = a1 if a1 > q1[1] else q1[1]

            delta = p1 - p0
            p0 -= 0.05 * delta
            p1 += 0.05 * delta

            x = np.linspace(p0[0], p1[0], Nx)
            y = np.linspace(p0[1], p1[1], Ny)

            pts = np.zeros((Nx, Ny, 2))
            pts[..., 0] = x[:, None]
            pts[..., 1] = y[None, :]
            pts = pts.reshape(Nx * Ny, 2)

            pars = dict(shading="auto", cmap="magma_r", zorder=-100)
            sbar = True
            if fill_mode == "dist":
                dists = self.points_distance(pts).reshape(Nx, Ny)
            elif fill_mode == "dist_inside":
                ins = self.points_inside(pts)
                dists = np.full(Nx * Ny, np.nan, dtype=np.float64)
                dists[ins] = self.points_distance(pts[ins])
                dists = dists.reshape(Nx, Ny)
            elif fill_mode[:7] == "inside_":
                ins = self.points_inside(pts)
                dists = np.full(Nx * Ny, np.nan, dtype=np.float64)
                dists[ins] = 1.0
                dists = dists.reshape(Nx, Ny)
                pars["cmap"] = ListedColormap([fill_mode[7:]])
                sbar = False
            elif fill_mode == "dist_outside":
                ins = self.points_inside(pts)
                dists = np.full(Nx * Ny, np.nan, dtype=np.float64)
                dists[~ins] = self.points_distance(pts[~ins])
                dists = dists.reshape(Nx, Ny)
            elif fill_mode[:8] == "outside_":
                ins = self.points_inside(pts)
                dists = np.full(Nx * Ny, np.nan, dtype=np.float64)
                dists[~ins] = 1
                dists = dists.reshape(Nx, Ny)
                pars["cmap"] = ListedColormap([fill_mode[8:]])
                sbar = False
            else:
                raise ValueError(
                    f"Illegal parameter 'fill_mode = {fill_mode}', expecting: None, dist, dist_inside, dist_outside"
                )

            pars.update(pars_distance)
            im = ax.pcolormesh(x, y, dists.T, **pars)
            if sbar:
                plt.colorbar(im, ax=ax, orientation="vertical", label="distance")

        ax.autoscale_view()
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")
        ax.set_aspect("equal", adjustable="box")

    def inverse(self):
        """
        Get the inverted geometry

        Returns
        -------
        inverted: foxes.utils.geom2d.InvertedAreaGeometry
            The inverted geometry

        """
        return InvertedAreaGeometry(self)

    def __add__(self, g):
        if isinstance(g, list):
            return AreaUnion([self] + g)
        elif isinstance(g, AreaUnion):
            return AreaUnion([self] + g.geometries)
        else:
            return AreaUnion([self, g])

    def __sub__(self, g):
        if isinstance(g, list):
            return AreaIntersection([self] + g.inverse())
        else:
            return AreaIntersection([self, g.inverse()])


class InvertedAreaGeometry(AreaGeometry):
    """
    Base class for inverted geometries.

    :group: utils.geom2d

    """

    def __init__(self, geometry):
        """
        Constructor.

        Parameters
        ----------
        geometry: geom2d.AreaGeometry
            The original geometry

        """
        self._geometry = geometry

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        pmi = self._geometry.p_min()
        if not np.any(np.isinf(pmi)):
            return np.full(2, -np.inf, dtype=np.float64)
        elif isinstance(self._geometry, InvertedAreaGeometry):
            out = np.full(2, np.inf, dtype=np.float64)
            imi = self._geometry.inverse().p_min()
            for di in range(2):
                if np.isinf(pmi[di]) and not np.isinf(imi[di]):
                    out[di] = np.minimum(out[di], imi[di])
                if not np.isinf(pmi[di]):
                    out[di] = -np.inf
            return out
        else:
            return np.full(2, -np.inf, dtype=np.float64)

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        pma = self._geometry.p_max()
        if not np.any(np.isinf(pma)):
            return np.full(2, np.inf, dtype=np.float64)
        elif isinstance(self._geometry, InvertedAreaGeometry):
            out = np.full(2, -np.inf, dtype=np.float64)
            ima = self._geometry.inverse().p_max()
            for di in range(2):
                if np.isinf(pma[di]) and not np.isinf(ima[di]):
                    out[di] = np.maximum(out[di], ima[di])
                if not np.isinf(pma[di]):
                    out[di] = np.inf
            return out
        else:
            return np.full(2, np.inf, dtype=np.float64)

    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """
        return self._geometry.points_distance(points, return_nearest)

    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        return ~self._geometry.points_inside(points)

    def add_to_figure(
        self,
        ax,
        show_boundary=False,
        fill_mode="inside_slategray",
        pars_boundary={},
        pars_distance={},
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        self._geometry.add_to_figure(
            ax,
            show_boundary,
            fill_mode=None,
            pars_boundary=pars_boundary,
            pars_distance={},
        )
        super().add_to_figure(
            ax, show_boundary, fill_mode, pars_boundary, pars_distance
        )

    def inverse(self):
        """
        Get the inverted geometry

        Returns
        -------
        inverted: foxes.utils.geom2d.InvertedAreaGeometry
            The inverted geometry

        """
        return self._geometry


class AreaUnion(AreaGeometry):
    """
    The union of area geometries.

    Attributes
    ----------
    geometries: list of geom2d.AreaGeometry
        The geometries

    :group: utils.geom2d

    """

    def __init__(self, geometries):
        """
        Constructor.

        Parameters
        ----------
        geometries: list of geom2d.AreaGeometry
            The geometries

        """
        self.geometries = geometries

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        out = None
        for g in self.geometries:
            pmi = g.p_min()
            if out is None:
                out = pmi
            else:
                out = np.minimum(out, pmi)
        return out

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        out = None
        for g in self.geometries:
            pma = g.p_max()
            if out is None:
                out = pma
            else:
                out = np.maximum(out, pma)
        return out

    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """
        if len(self.geometries) == 1:
            return self.geometries[0].points_distance(points, return_nearest)

        n_pts = len(points)
        dist = np.full(n_pts, np.inf, dtype=np.float64)
        pins = np.zeros(n_pts, dtype=bool)
        nerst = np.zeros((n_pts, 2), dtype=np.float64) if return_nearest else None
        for g in self.geometries:
            res = g.points_distance(points, return_nearest)
            ins = g.points_inside(points)
            d = res[0] if return_nearest else res

            # was outside, is outside:
            sel = ~pins & ~ins & (d < dist)
            if np.any(sel):
                dist[sel] = d[sel]
                if return_nearest:
                    nerst[sel] = res[1][sel]

            # was outside, is inside:
            sel = ~pins & ins
            if np.any(sel):
                pins[sel] = True
                dist[sel] = d[sel]
                if return_nearest:
                    nerst[sel] = res[1][sel]

            # was inside, is inside:
            sel = pins & ins & (d > dist)
            if np.any(sel):
                dist[sel] = d[sel]
                if return_nearest:
                    nerst[sel] = res[1][sel]

        if return_nearest:
            return dist, nerst
        else:
            return dist

    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        if len(self.geometries) == 1:
            return self.geometries[0].points_inside(points)

        n_pts = len(points)
        inside = np.zeros(n_pts, dtype=bool)
        for g in self.geometries:
            inside = inside | g.points_inside(points)
        return inside

    def add_to_figure(
        self,
        ax,
        show_boundary=False,
        fill_mode="inside_slategray",
        pars_boundary={},
        pars_distance={},
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        if show_boundary:
            for g in self.geometries:
                g.add_to_figure(
                    ax,
                    show_boundary=True,
                    fill_mode=None,
                    pars_boundary=pars_boundary,
                    pars_distance={},
                )

        super().add_to_figure(
            ax,
            show_boundary=False,
            fill_mode=fill_mode,
            pars_boundary={},
            pars_distance=pars_distance,
        )

    def inverse(self):
        """
        Get the inverted geometry

        Returns
        -------
        inverted: foxes.utils.geom2d.InvertedAreaGeometry
            The inverted geometry

        """
        return InvertedAreaUnion(self)

    def __add__(self, g):
        if isinstance(g, list):
            return AreaUnion(self.geometries + g)
        elif isinstance(g, AreaUnion):
            return AreaUnion(self.geometries + g.geometries)
        else:
            return AreaUnion(self.geometries + [g])


class InvertedAreaUnion(InvertedAreaGeometry):
    """
    Inversion of a union of areas

    :group: utils.geom2d

    """

    def __init__(self, union):
        """
        Constructor.

        Parameters
        ----------
        union: geom2d.AreaUnion
            The original area union geometry

        """
        super().__init__(union)

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        if len(self._geometry.geometries) == 1:
            return self._geometry.geometries[0].inverse().p_min()

        pmi = self._geometry.p_min()
        if not np.any(np.isinf(pmi)):
            return np.full(2, -np.inf, dtype=np.float64)
        else:
            out = np.full(2, np.inf, dtype=np.float64)
            for g in self._geometry.geometries:
                imi = g.inverse().p_min()
                for di in range(2):
                    if np.isinf(pmi[di]) and not np.isinf(imi[di]):
                        out[di] = np.minimum(out[di], imi[di])
            for di in range(2):
                if not np.isinf(pmi[di]):
                    out[di] = -np.inf
            return out

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        if len(self._geometry.geometries) == 1:
            return self._geometry.geometries[0].inverse().p_max()

        pma = self._geometry.p_max()
        if not np.any(np.isinf(pma)):
            return np.full(2, np.inf, dtype=np.float64)
        else:
            out = np.full(2, -np.inf, dtype=np.float64)
            for g in self._geometry.geometries:
                ima = g.inverse().p_max()
                for di in range(2):
                    if np.isinf(pma[di]) and not np.isinf(ima[di]):
                        out[di] = np.maximum(out[di], ima[di])
            for di in range(2):
                if not np.isinf(pma[di]):
                    out[di] = np.inf
            return out


class AreaIntersection(AreaGeometry):
    """
    The intersection of area geometries.

    :group: utils.geom2d

    """

    def __new__(cls, geometries):
        """
        Constructor.

        Parameters
        ----------
        geometries: list of geom2d.AreaGeometry
            The geometries

        """
        return AreaUnion([g.inverse() for g in geometries]).inverse()



class WindFarm:
    """
    The wind farm.

    Attributes
    ----------
    name: str
        The wind farm name
    turbines: list of foxes.core.Turbine
        The wind turbines
    boundary: foxes.utils.geom2d.AreaGeometry, optional
        The wind farm boundary

    :group: core

    """

    def __init__(self, name="wind_farm", boundary=None):
        """
        Constructor.

        Parameters
        ----------
        name: str
            The wind farm name

        """
        self.name = name
        self.turbines = []
        self.boundary = boundary

    def add_turbine(self, turbine, verbosity=1):
        """
        Add a wind turbine to the list.

        Parameters
        ----------
        turbine: foxes.core.Turbine
            The wind turbine
        verbosity: int
            The output verbosity, 0 = silent

        """
        if turbine.index is None:
            turbine.index = len(self.turbines)
        if turbine.name is None:
            turbine.name = f"T{turbine.index}"
        self.turbines.append(turbine)
        if verbosity > 0:
            print(
                f"Turbine {turbine.index}, {turbine.name}: xy=({turbine.xy[0]:.2f}, {turbine.xy[1]:.2f}), {', '.join(turbine.models)}"
            )

    @property
    def n_turbines(self):
        """
        The number of turbines in the wind farm

        Returns
        -------
        n_turbines: int
            The total number of turbines

        """
        return len(self.turbines)

    @property
    def turbine_names(self):
        """
        The list of names of all turbines

        Returns
        -------
        names: list of str
            The names of all turbines

        """
        return [t.name for t in self.turbines]



"""
Turbine type models.
"""

from .PCt_file import PCtFile
from .PCt_from_two import PCtFromTwo
from .CpCt_file import CpCtFile
from .CpCt_from_two import CpCtFromTwo
from .null_type import NullType
from .wsrho2PCt_from_two import WsRho2PCtFromTwo
from .wsti2PCt_from_two import WsTI2PCtFromTwo
